from functools import partial
import io
import json
import platform

import pytest

import falcon
from falcon import media
from falcon import testing
import falcon.asgi
from falcon.asgi.stream import BoundedStream
from falcon.constants import MEDIA_JSON
from falcon.constants import MEDIA_YAML

msgspec = None
mujson = None
orjson = None
rapidjson = None
ujson = None

try:
    import msgspec  # type: ignore
except ImportError:
    pass

try:
    import mujson  # type: ignore
except ImportError:
    pass

try:
    import rapidjson  # type: ignore
except ImportError:
    pass

try:
    import ujson  # type: ignore
except ImportError:
    pass

if platform.python_implementation() == 'CPython':
    try:
        import orjson  # type: ignore
    except ImportError:
        pass

YEN = b'\xc2\xa5'

SERIALIZATION_PARAM_LIST = [
    # Default json.dumps, with only ascii
    (None, {'test': 'value'}, b'{"test":"value"}'),
    (
        partial(lambda media, **kwargs: json.dumps([media, kwargs]), ensure_ascii=True),
        {'test': 'value'},
        b'[{"test":"value"},{"ensure_ascii":true}]',
    ),
    # Default json.dumps, with non-ascii characters
    (None, {'yen': YEN.decode()}, b'{"yen":"' + YEN + b'"}'),
]

DESERIALIZATION_PARAM_LIST = [
    (None, b'[1, 2]', [1, 2]),
    (
        partial(
            json.loads, object_hook=lambda data: {k: v.upper() for k, v in data.items()}
        ),
        b'{"key": "value"}',
        {'key': 'VALUE'},
    ),
]

ALL_JSON_IMPL = [(json.dumps, json.loads)]
ALL_JSON_IMPL_IDS = ['stdlib']


if msgspec:
    SERIALIZATION_PARAM_LIST += [
        (
            msgspec.json.encode,
            {'test': 'value'},
            b'{"test":"value"}',
        ),
    ]
    DESERIALIZATION_PARAM_LIST += [
        (msgspec.json.decode, b'{"test": "value"}', {'test': 'value'}),
    ]
    ALL_JSON_IMPL += [(msgspec.json.encode, msgspec.json.decode)]
    ALL_JSON_IMPL_IDS += ['msgspec']

if mujson:
    SERIALIZATION_PARAM_LIST += [
        (
            partial(mujson.dumps, ensure_ascii=True),
            {'test': 'value'},
            b'{"test":"value"}',
        ),
    ]
    DESERIALIZATION_PARAM_LIST += [
        (mujson.loads, b'{"test": "value"}', {'test': 'value'}),
    ]
    ALL_JSON_IMPL += [(partial(mujson.dumps, ensure_ascii=True), mujson.loads)]
    ALL_JSON_IMPL_IDS += ['mujson']

if orjson:
    SERIALIZATION_PARAM_LIST += [
        (orjson.dumps, {'test': 'value'}, b'{"test":"value"}'),
    ]
    DESERIALIZATION_PARAM_LIST += [
        (orjson.loads, b'{"test": "value"}', {'test': 'value'}),
    ]
    ALL_JSON_IMPL += [(orjson.dumps, orjson.loads)]
    ALL_JSON_IMPL_IDS += ['orjson']

if rapidjson:
    SERIALIZATION_PARAM_LIST += [
        (rapidjson.dumps, {'test': 'value'}, b'{"test":"value"}'),
    ]
    DESERIALIZATION_PARAM_LIST += [
        (rapidjson.loads, b'{"test": "value"}', {'test': 'value'}),
    ]
    ALL_JSON_IMPL += [(rapidjson.dumps, rapidjson.loads)]
    ALL_JSON_IMPL_IDS += ['rapidjson']

if ujson:
    SERIALIZATION_PARAM_LIST += [
        (ujson.dumps, {'test': 'value'}, b'{"test":"value"}'),
    ]
    DESERIALIZATION_PARAM_LIST += [
        (ujson.loads, b'{"test": "value"}', {'test': 'value'}),
    ]
    ALL_JSON_IMPL += [(ujson.dumps, ujson.loads)]
    ALL_JSON_IMPL_IDS += ['ujson']


@pytest.mark.parametrize(
    'library, name',
    [
        (msgspec, 'msgspec'),
        (mujson, 'mujson'),
        (orjson, 'orjson'),
        (rapidjson, 'rapidjson'),
        (ujson, 'ujson'),
    ],
    ids=['msgspec', 'mujson', 'orjson', 'rapidjson', 'ujson'],
)
def test_check_json_library(library, name):
    # NOTE(vytas): A synthetic test just to visualize which JSON libraries
    #   are absent and skipped.
    if library is None:
        pytest.skip(f'{name} is not installed')

    # NOTE(vytas): msgspec does not employ the names dumps() & loads().
    assert hasattr(library, 'dumps') or hasattr(library, 'json')
    assert hasattr(library, 'loads') or hasattr(library, 'json')


@pytest.mark.parametrize('func, body, expected', SERIALIZATION_PARAM_LIST)
def test_serialization(asgi, func, body, expected):
    handler = media.JSONHandler(dumps=func)

    args = (body, b'application/javacript')

    if asgi:
        result = falcon.async_to_sync(handler.serialize_async, *args)
    else:
        result = handler.serialize(*args)

    # NOTE(nZac) PyPy and CPython render the final string differently. One
    # includes spaces and the other doesn't. This replace will normalize that.
    assert result.replace(b' ', b'') == expected  # noqa


@pytest.mark.parametrize('func, body, expected', DESERIALIZATION_PARAM_LIST)
def test_deserialization(asgi, func, body, expected):
    handler = media.JSONHandler(loads=func)

    args = ['application/javacript', len(body)]

    if asgi:
        s = BoundedStream(testing.ASGIRequestEventEmitter(body))
        args.insert(0, s)

        result = falcon.async_to_sync(handler.deserialize_async, *args)
    else:
        args.insert(0, io.BytesIO(body))
        result = handler.deserialize(*args)

    assert result == expected


@pytest.mark.parametrize('dumps, loads', ALL_JSON_IMPL, ids=ALL_JSON_IMPL_IDS)
@pytest.mark.parametrize('subclass', (True, False))
def test_full_app(asgi, util, dumps, loads, subclass):
    if subclass:

        class JSONHandlerSubclass(media.JSONHandler):
            pass

        handler = JSONHandlerSubclass(dumps=dumps, loads=loads)
        assert handler._serialize_sync is None
        assert handler._deserialize_sync is None
    else:
        handler = media.JSONHandler(dumps=dumps, loads=loads)
        assert handler._serialize_sync is not None
        assert handler._deserialize_sync is not None
    app = util.create_app(asgi)
    app.req_options.media_handlers[falcon.MEDIA_JSON] = handler
    app.resp_options.media_handlers[falcon.MEDIA_JSON] = handler

    if asgi:

        class Resp:
            async def on_get(self, req, res):
                res.media = await req.get_media()

    else:

        class Resp:
            def on_get(self, req, res):
                res.media = req.get_media()

    app.add_route('/go', Resp())

    data = {'foo': 123, 'bar': [2, 3], 'baz': {'x': 'y'}}
    res = testing.simulate_get(app, '/go', json=data)
    assert res.json == data


@pytest.mark.parametrize(
    'handler_mt',
    [
        'application/json',
        # NOTE(kgriffs): Include a bogus parameter to validate fuzzy matching logic
        'application/json; answer=42',
    ],
)
def test_deserialization_raises(asgi, util, handler_mt):
    app = util.create_app(asgi)

    class SuchException(Exception):
        pass

    class FaultyHandler(media.BaseHandler):
        def deserialize(self, stream, content_type, content_length):
            raise SuchException('Wow such error.')

        def deserialize_async(self, stream, content_type, content_length):
            raise SuchException('Wow such error.')

        def serialize(self, media, content_type):
            raise SuchException('Wow such error.')

    handlers = media.Handlers({handler_mt: FaultyHandler()})

    app.req_options.media_handlers = handlers
    app.resp_options.media_handlers = handlers

    class Resource:
        def on_get(self, req, resp):
            resp.media = {}

        def on_post(self, req, resp):
            req.media

    class ResourceAsync:
        async def on_get(self, req, resp):
            resp.media = {}

        async def on_post(self, req, resp):
            # NOTE(kgriffs): In this one case we use the property
            #   instead of get_media(), in order to test that the
            #   alias works as expected.
            await req.media

    app.add_route('/', ResourceAsync() if asgi else Resource())

    # NOTE(kgriffs): Now that we install a default handler for
    #   Exception, we have to clear them to test the path we want
    #   to trigger.
    # TODO(kgriffs): Since we always add a default error handler
    #   for Exception, should we take out the checks in the WSGI/ASGI
    #   callable and just always assume it will be handled? If so,
    #   it makes testing that the right exception is propagated harder;
    #   I suppose we'll have to look at what is logged.
    app._error_handlers.clear()

    with pytest.raises(SuchException):
        testing.simulate_get(app, '/')

    with pytest.raises(SuchException):
        testing.simulate_post(app, '/', json={})


def test_nonstandard_deserialization_error(asgi, util):
    # NOTE(vytas): At the time of writing, this is also tested by msgspec.
    #   However, we add an additional generic test here if msgspec restructures
    #   its class hierarchy to inherit from ValueError, or if it is unavailable
    #   in the environment in question.

    class SuchException(RuntimeError):
        def __init__(self):
            super().__init__('wow such exception')

    def fancy_loads(data):
        try:
            return json.loads(data)
        except ValueError:
            raise SuchException()

    class DevNull:
        def on_post(self, req, resp):
            req.get_media()

        async def on_post_async(self, req, resp):
            await req.get_media()

    app = util.create_app(asgi)
    app.add_route('/dev/null', DevNull(), suffix=('async' if asgi else None))
    app.req_options.media_handlers[falcon.MEDIA_JSON] = media.JSONHandler(
        loads=fancy_loads
    )

    resp = testing.simulate_post(
        app, '/dev/null', body=b'Hello: world', content_type=falcon.MEDIA_JSON
    )
    assert resp.status_code == 400
    assert resp.json.get('title') == 'Invalid JSON'


@pytest.mark.parametrize(
    'custom_json_handler', [True, False], ids=('CustomJSONHandler', 'JSONHandler')
)
@pytest.mark.parametrize(
    'custom_resp_type', [True, False], ids=('CustomResponse', 'Response')
)
def test_dumps_bytes_output(asgi, util, custom_json_handler, custom_resp_type):
    # NOTE(vytas): This scenario is already exercised by certain third party
    #   serialization libraries (orjson,  msgspec).
    #   However, they might be unavailable on a new version of CPython, or when
    #   running minimal tests, so we cover the relevant code paths with this
    #   test case without any external dependencies.

    def dumpb(obj, **kwargs):
        return json.dumps(obj, **kwargs).encode()

    class Resource:
        def on_get(self, req, resp):
            resp.media = {'msg': 'Hello'}

        async def on_get_async(self, req, resp):
            resp.media = {'msg': 'Hello'}

    json_handler_type = media.JSONHandler
    if custom_json_handler:
        json_handler_type = type('CustomJSONHandler', (media.JSONHandler,), {})

    response_type = None
    standard_response_type = falcon.asgi.Response if asgi else falcon.Response
    if custom_resp_type:
        response_type = type('CustomResponse', (standard_response_type,), {})

    app = util.create_app(asgi, response_type=response_type)
    app.add_route('/msg', Resource(), suffix=('async' if asgi else None))
    app.resp_options.media_handlers[falcon.MEDIA_JSON] = json_handler_type(dumps=dumpb)

    resp = testing.simulate_get(app, '/msg')
    assert resp.status_code == 200
    assert resp.json == {'msg': 'Hello'}


def test_sync_methods_not_overridden(asgi, util):
    app = util.create_app(asgi)

    class FaultyHandler(media.BaseHandler):
        pass

    handlers = media.Handlers({'application/json': FaultyHandler()})
    app.req_options.media_handlers = handlers
    app.resp_options.media_handlers = handlers

    class Resource:
        def on_get(self, req, resp):
            resp.media = {}

        def on_post(self, req, resp):
            req.media

    class ResourceAsync:
        async def on_get(self, req, resp):
            resp.media = {}

        async def on_post(self, req, resp):
            await req.get_media()

    app.add_route('/', ResourceAsync() if asgi else Resource())

    # NOTE(caselit): force serialization in xml,
    # since error.to_json uses the faulty handler
    result = testing.simulate_get(app, '/', headers={'Accept': 'text/xml'})
    assert result.status_code == 500

    result = testing.simulate_post(app, '/', json={}, headers={'Accept': 'text/xml'})
    assert result.status_code == 500


def test_async_methods_not_overridden(util):
    app = util.create_app(asgi=True)

    class SimpleHandler(media.BaseHandler):
        def serialize(self, media, content_type):
            return json.dumps(media).encode()

        def deserialize(self, stream, content_type, content_length):
            return json.load(stream)

    handlers = media.Handlers({'application/json': SimpleHandler()})
    app.req_options.media_handlers = handlers
    app.resp_options.media_handlers = handlers

    class ResourceAsync:
        async def on_post(self, req, resp):
            resp.media = await req.get_media()

    app.add_route('/', ResourceAsync())

    doc = {'event': 'serialized'}
    result = testing.simulate_post(app, '/', json=doc)
    assert result.status_code == 200
    assert result.json == doc


def test_async_handler_returning_none(util):
    app = util.create_app(asgi=True)

    class SimpleHandler(media.BaseHandler):
        def serialize(self, media, content_type):
            return json.dumps(media).encode()

        def deserialize(self, stream, content_type, content_length):
            return None

    handlers = media.Handlers({'application/json': SimpleHandler()})
    app.req_options.media_handlers = handlers
    app.resp_options.media_handlers = handlers

    class ResourceAsync:
        async def on_post(self, req, resp):
            resp.media = [await req.get_media()]

    app.add_route('/', ResourceAsync())

    doc = {'event': 'serialized'}
    result = testing.simulate_post(app, '/', json=doc)
    assert result.status_code == 200
    assert result.json == [None]


def test_json_err_no_handler(asgi, util):
    app = util.create_app(asgi)

    handlers = media.Handlers({falcon.MEDIA_URLENCODED: media.URLEncodedFormHandler()})

    app.req_options.media_handlers = handlers
    app.resp_options.media_handlers = handlers

    class Resource:
        def on_get(self, req, resp):
            raise falcon.HTTPForbidden()

    app.add_route('/', Resource())

    result = testing.simulate_get(app, '/')
    assert result.status_code == 403
    assert result.json == falcon.HTTPForbidden().to_dict()


def test_handlers_include_new_media_handlers_in_resolving() -> None:
    handlers = media.Handlers({falcon.MEDIA_URLENCODED: media.URLEncodedFormHandler()})
    assert handlers._resolve(MEDIA_YAML, MEDIA_JSON, raise_not_found=False)[0] is None

    js_handler = media.JSONHandler()
    handlers[MEDIA_YAML] = js_handler
    resolved = handlers._resolve(MEDIA_YAML, MEDIA_JSON, raise_not_found=False)[0]
    assert resolved is js_handler


class TestBaseHandler:
    def test_defaultError(self):
        h = media.BaseHandler()

        def test(call):
            with pytest.raises(NotImplementedError) as e:
                call()

            assert e.value.args == ()

        test(lambda: h.serialize({}, 'my-type'))
        test(lambda: h.deserialize('', 'my-type', 0))

    def test_json(self):
        h = media.BaseHandler()

        with pytest.raises(
            NotImplementedError, match='The JSON media handler requires'
        ):
            h.serialize({}, falcon.MEDIA_JSON)
        with pytest.raises(
            NotImplementedError, match='The JSON media handler requires'
        ):
            h.deserialize('', falcon.MEDIA_JSON, 0)

        with pytest.raises(
            NotImplementedError, match='The JSON media handler requires'
        ):
            h.serialize({}, falcon.MEDIA_JSON + '; charset=UTF-8')
        with pytest.raises(
            NotImplementedError, match='The JSON media handler requires'
        ):
            h.deserialize('', falcon.MEDIA_JSON + '; charset=UTF-8', 0)



# Copyright 2013 by Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for the App class."""

from __future__ import annotations

from inspect import iscoroutinefunction
from typing import IO, Iterable, List, Literal, Optional, overload, Tuple, Union

from falcon import util
from falcon._typing import AsgiProcessRequestMethod as APRequest
from falcon._typing import AsgiProcessRequestWsMethod
from falcon._typing import AsgiProcessResourceMethod as APResource
from falcon._typing import AsgiProcessResourceWsMethod
from falcon._typing import AsgiProcessResponseMethod as APResponse
from falcon._typing import AsyncMiddleware
from falcon._typing import ProcessRequestMethod as PRequest
from falcon._typing import ProcessResourceMethod as PResource
from falcon._typing import ProcessResponseMethod as PResponse
from falcon._typing import SyncMiddleware
from falcon.constants import MEDIA_JSON
from falcon.constants import MEDIA_XML
from falcon.errors import CompatibilityError
from falcon.errors import HTTPError
from falcon.request import Request
from falcon.response import Response
from falcon.util.sync import _wrap_non_coroutine_unsafe

__all__ = (
    'prepare_middleware',
    'prepare_middleware_ws',
    'default_serialize_error',
    'CloseableStreamIterator',
)

PreparedMiddlewareResult = Tuple[
    Union[
        Tuple[PRequest, ...], Tuple[Tuple[Optional[PRequest], Optional[PResource]], ...]
    ],
    Tuple[PResource, ...],
    Tuple[PResponse, ...],
]
AsyncPreparedMiddlewareResult = Tuple[
    Union[
        Tuple[APRequest, ...],
        Tuple[Tuple[Optional[APRequest], Optional[APResource]], ...],
    ],
    Tuple[APResource, ...],
    Tuple[APResponse, ...],
]


@overload
def prepare_middleware(
    middleware: Iterable[SyncMiddleware],
    independent_middleware: bool = ...,
    asgi: Literal[False] = ...,
) -> PreparedMiddlewareResult: ...


@overload
def prepare_middleware(
    middleware: Iterable[AsyncMiddleware],
    independent_middleware: bool = ...,
    *,
    asgi: Literal[True],
) -> AsyncPreparedMiddlewareResult: ...


@overload
def prepare_middleware(
    middleware: Union[Iterable[SyncMiddleware], Iterable[AsyncMiddleware]],
    independent_middleware: bool = ...,
    asgi: bool = ...,
) -> Union[PreparedMiddlewareResult, AsyncPreparedMiddlewareResult]: ...


def prepare_middleware(
    middleware: Union[Iterable[SyncMiddleware], Iterable[AsyncMiddleware]],
    independent_middleware: bool = False,
    asgi: bool = False,
) -> Union[PreparedMiddlewareResult, AsyncPreparedMiddlewareResult]:
    """Check middleware interfaces and prepare the methods for request handling.

    Note:
        This method is only applicable to WSGI apps.

    Arguments:
        middleware (iterable): An iterable of middleware objects.

    Keyword Args:
        independent_middleware (bool): ``True`` if the request and
            response middleware methods should be treated independently
            (default ``False``)
        asgi (bool): ``True`` if an ASGI app, ``False`` otherwise
            (default ``False``)

    Returns:
        tuple: A tuple of prepared middleware method tuples
    """

    # PERF(kgriffs): do getattr calls once, in advance, so we don't
    # have to do them every time in the request path.
    request_mw: Union[
        List[PRequest],
        List[Tuple[Optional[PRequest], Optional[PResource]]],
        List[APRequest],
        List[Tuple[Optional[APRequest], Optional[APResource]]],
    ] = []
    resource_mw: Union[List[APResource], List[PResource]] = []
    response_mw: Union[List[APResponse], List[PResponse]] = []

    for component in middleware:
        # NOTE(kgriffs): Middleware that supports both WSGI and ASGI can
        #   append an *_async postfix to the ASGI version of the method
        #   to distinguish the two. Otherwise, the prefix is unnecessary.

        if asgi:
            process_request: Union[Optional[APRequest], Optional[PRequest]] = (
                util.get_bound_method(component, 'process_request_async')
                or _wrap_non_coroutine_unsafe(
                    util.get_bound_method(component, 'process_request')
                )
            )

            process_resource: Union[Optional[APResource], Optional[PResource]] = (
                util.get_bound_method(component, 'process_resource_async')
                or _wrap_non_coroutine_unsafe(
                    util.get_bound_method(component, 'process_resource')
                )
            )

            process_response: Union[Optional[APResponse], Optional[PResponse]] = (
                util.get_bound_method(component, 'process_response_async')
                or _wrap_non_coroutine_unsafe(
                    util.get_bound_method(component, 'process_response')
                )
            )

            for m in (process_request, process_resource, process_response):
                # NOTE(kgriffs): iscoroutinefunction() always returns False
                #   for cythonized functions.
                #
                #   https://github.com/cython/cython/issues/2273
                #   https://bugs.python.org/issue38225
                #
                if m and not iscoroutinefunction(m) and util.is_python_func(m):
                    msg = (
                        '{} must be implemented as an awaitable coroutine. If '
                        'you would like to retain compatibility '
                        'with WSGI apps, the coroutine versions of the '
                        'middleware methods may be implemented side-by-side '
                        'by applying an *_async postfix to the method names. '
                    )
                    raise CompatibilityError(msg.format(m))

        else:
            process_request = util.get_bound_method(component, 'process_request')
            process_resource = util.get_bound_method(component, 'process_resource')
            process_response = util.get_bound_method(component, 'process_response')

            for m in (process_request, process_resource, process_response):
                if m and iscoroutinefunction(m):
                    msg = (
                        '{} may not implement coroutine methods and '
                        'remain compatible with WSGI apps without '
                        'using the *_async postfix to explicitly identify '
                        'the coroutine version of a given middleware '
                        'method.'
                    )
                    raise CompatibilityError(msg.format(component))

        if not (process_request or process_resource or process_response):
            if asgi and any(
                hasattr(component, m)
                for m in [
                    'process_startup',
                    'process_shutdown',
                    'process_request_ws',
                    'process_resource_ws',
                ]
            ):
                # NOTE(kgriffs): This middleware only has ASGI lifespan
                #   event handlers
                continue

            msg = '{0} must implement at least one middleware method'
            raise TypeError(msg.format(component))

        # NOTE: depending on whether we want to execute middleware
        # independently, we group response and request middleware either
        # together or separately.
        if independent_middleware:
            if process_request:
                request_mw.append(process_request)  # type: ignore[arg-type]
            if process_response:
                response_mw.insert(0, process_response)  # type: ignore[arg-type]
        else:
            if process_request or process_response:
                request_mw.append((process_request, process_response))  # type: ignore[arg-type]

        if process_resource:
            resource_mw.append(process_resource)  # type: ignore[arg-type]

    return tuple(request_mw), tuple(resource_mw), tuple(response_mw)  # type: ignore[return-value]


AsyncPreparedMiddlewareWsResult = Tuple[
    Tuple[AsgiProcessRequestWsMethod, ...], Tuple[AsgiProcessResourceWsMethod, ...]
]


def prepare_middleware_ws(
    middleware: Iterable[AsyncMiddleware],
) -> AsyncPreparedMiddlewareWsResult:
    """Check middleware interfaces and prepare WebSocket methods for request handling.

    Note:
        This method is only applicable to ASGI apps.

    Arguments:
        middleware (iterable): An iterable of middleware objects.

    Returns:
        tuple: A two-item ``(request_mw, resource_mw)`` tuple, where
        *request_mw* is an ordered list of ``process_request_ws()`` methods,
        and *resource_mw* is an ordered list of ``process_resource_ws()``
        methods.
    """

    # PERF(kgriffs): do getattr calls once, in advance, so we don't
    # have to do them every time in the request path.
    request_mw: List[AsgiProcessRequestWsMethod] = []
    resource_mw: List[AsgiProcessResourceWsMethod] = []

    process_request_ws: Optional[AsgiProcessRequestWsMethod]
    process_resource_ws: Optional[AsgiProcessResourceWsMethod]

    for component in middleware:
        process_request_ws = util.get_bound_method(component, 'process_request_ws')
        process_resource_ws = util.get_bound_method(component, 'process_resource_ws')

        for m in (process_request_ws, process_resource_ws):
            if not m:
                continue

            # NOTE(kgriffs): iscoroutinefunction() always returns False
            #   for cythonized functions.
            #
            #   https://github.com/cython/cython/issues/2273
            #   https://bugs.python.org/issue38225
            #
            if not iscoroutinefunction(m) and util.is_python_func(m):
                msg = '{} must be implemented as an awaitable coroutine.'
                raise CompatibilityError(msg.format(m))

        if process_request_ws:
            request_mw.append(process_request_ws)

        if process_resource_ws:
            resource_mw.append(process_resource_ws)

    return tuple(request_mw), tuple(resource_mw)


def default_serialize_error(req: Request, resp: Response, exception: HTTPError) -> None:
    """Serialize the given instance of HTTPError.

    This function determines which of the supported media types, if
    any, are acceptable by the client, and serializes the error
    to the preferred type.

    Currently, JSON and XML are the only supported media types. If the
    client accepts both JSON and XML with equal weight, JSON will be
    chosen.

    Other media types can be supported by using a custom error serializer.

    Note:
        If a custom media type is used and the type includes a
        "+json" or "+xml" suffix, the error will be serialized
        to JSON or XML, respectively. If this behavior is not
        desirable, a custom error serializer may be used to
        override this one.

    Args:
        req: Instance of ``falcon.Request``
        resp: Instance of ``falcon.Response``
        exception: Instance of ``falcon.HTTPError``
    """
    options = resp.options
    predefined = (
        [MEDIA_JSON, 'text/xml', MEDIA_XML]
        if options.xml_error_serialization
        else [MEDIA_JSON]
    )
    media_handlers = [mt for mt in options.media_handlers if mt not in predefined]
    # NOTE(caselit,vytas): Add the registered handlers after the predefined
    #   ones. This ensures that in the case of an equal match, the first one
    #   (JSON) is selected and that the q parameter is taken into consideration
    #   when selecting the media handler.
    preferred = req.client_prefers(predefined + media_handlers)

    if preferred is None:
        # NOTE(kgriffs): See if the client expects a custom media
        # type based on something Falcon supports. Returning something
        # is probably better than nothing, but if that is not
        # desired, this behavior can be customized by adding a
        # custom HTTPError serializer for the custom type.
        accept = req.accept.lower()

        # NOTE(kgriffs): Simple heuristic, but it's fast, and
        # should be sufficiently accurate for our purposes. Does
        # not take into account weights if both types are
        # acceptable (simply chooses JSON). If it turns out we
        # need to be more sophisticated, we can always change it
        # later (YAGNI).
        if '+json' in accept:
            preferred = MEDIA_JSON
        elif '+xml' in accept:
            # NOTE(caselit): Ignore xml_error_serialization when
            #   checking if the media should be XML. This gives a chance to
            #   an XML media handler, if any, to be used.
            preferred = MEDIA_XML

    if preferred is not None:
        handler, _, _ = options.media_handlers._resolve(
            preferred, MEDIA_JSON, raise_not_found=False
        )
        if preferred == MEDIA_JSON:
            # NOTE(caselit): Special case JSON to ensure that it's always
            #   possible to serialize an error in JSON even if no JSON handler
            #   is set in the media_handlers.
            resp.data = exception.to_json(handler)
        elif handler:
            # NOTE(caselit): Let the app serialize the response even if it
            #   needs to re-get the handler, since async handlers may not have
            #   a sync version available.
            resp.media = exception.to_dict()
        elif options.xml_error_serialization:
            resp.data = exception._to_xml()

        # NOTE(kgriffs): No need to append the charset param, since
        #   utf-8 is the default for both JSON and XML.
        resp.content_type = preferred

    resp.append_header('Vary', 'Accept')


class CloseableStreamIterator:
    """Iterator that wraps a file-like stream with support for close().

    This iterator can be used to read from an underlying file-like stream
    in block_size-chunks until the response from the stream is an empty
    byte string.

    This class is used to wrap WSGI response streams when a
    wsgi_file_wrapper is not provided by the server.  The fact that it
    also supports closing the underlying stream allows use of (e.g.)
    Python tempfile resources that would be deleted upon close.

    Args:
        stream (object): Readable file-like stream object.
        block_size (int): Number of bytes to read per iteration.
    """

    def __init__(self, stream: IO[bytes], block_size: int) -> None:
        self._stream = stream
        self._block_size = block_size

    def __iter__(self) -> CloseableStreamIterator:
        return self

    def __next__(self) -> bytes:
        data = self._stream.read(self._block_size)

        if data == b'':
            raise StopIteration
        else:
            return data

    def close(self) -> None:
        try:
            self._stream.close()
        except (AttributeError, TypeError):
            pass



from __future__ import annotations

from typing import Any, Callable, Optional, Protocol

from falcon import errors
from falcon.media.base import BaseHandler
from falcon.media.base import BinaryBaseHandlerWS
from falcon.typing import AsyncReadableIO
from falcon.typing import ReadableIO


class MessagePackHandler(BaseHandler):
    """Handler built using the :mod:`msgpack` module.

    This handler uses ``msgpack.unpackb()`` and ``msgpack.Packer().pack()``. The
    MessagePack ``bin`` type is used to distinguish between Unicode strings
    (of type ``str``) and byte strings (of type ``bytes``).

    This handler will raise a :class:`falcon.MediaNotFoundError` when attempting
    to parse an empty body; it will raise a :class:`falcon.MediaMalformedError`
    if an error happens while parsing the body.

    Note:
        This handler requires the extra ``msgpack`` package (version 0.5.2
        or higher), which must be installed in addition to ``falcon`` from
        PyPI:

        .. code::

            $ pip install msgpack
    """

    _pack: Callable[[Any], bytes]
    _unpackb: UnpackMethod

    def __init__(self) -> None:
        import msgpack

        packer = msgpack.Packer(autoreset=True, use_bin_type=True)
        self._pack = packer.pack
        self._unpackb = msgpack.unpackb

        # NOTE(kgriffs): To be safe, only enable the optimized protocol when
        #   not subclassed.
        if type(self) is MessagePackHandler:
            self._serialize_sync = self._pack  # type: ignore[assignment]
            self._deserialize_sync = self._deserialize

    def _deserialize(self, data: bytes) -> Any:
        if not data:
            raise errors.MediaNotFoundError('MessagePack')
        try:
            # NOTE(jmvrbanac): Using unpackb since we would need to manage
            # a buffer for Unpacker() which wouldn't gain us much.
            return self._unpackb(data, raw=False)
        except ValueError as err:
            raise errors.MediaMalformedError('MessagePack') from err

    def deserialize(
        self,
        stream: ReadableIO,
        content_type: Optional[str],
        content_length: Optional[int],
    ) -> Any:
        return self._deserialize(stream.read())

    async def deserialize_async(
        self,
        stream: AsyncReadableIO,
        content_type: Optional[str],
        content_length: Optional[int],
    ) -> Any:
        return self._deserialize(await stream.read())

    def serialize(self, media: Any, content_type: Optional[str]) -> bytes:
        return self._pack(media)

    async def serialize_async(self, media: Any, content_type: Optional[str]) -> bytes:
        return self._pack(media)


class MessagePackHandlerWS(BinaryBaseHandlerWS):
    """WebSocket media handler for de(serializing) MessagePack to/from BINARY payloads.

    This handler uses ``msgpack.unpackb()`` and ``msgpack.packb()``. The
    MessagePack ``bin`` type is used to distinguish between Unicode strings
    (of type ``str``) and byte strings (of type ``bytes``).

    Note:
        This handler requires the extra ``msgpack`` package (version 0.5.2
        or higher), which must be installed in addition to ``falcon`` from
        PyPI:

        .. code::

            $ pip install msgpack
    """

    __slots__ = ('msgpack', 'packer')

    _pack: Callable[[Any], bytes]
    _unpackb: UnpackMethod

    def __init__(self) -> None:
        import msgpack

        packer = msgpack.Packer(autoreset=True, use_bin_type=True)
        self._pack = packer.pack
        self._unpackb = msgpack.unpackb

    def serialize(self, media: object) -> bytes:
        return self._pack(media)

    def deserialize(self, payload: bytes) -> Any:
        # NOTE(jmvrbanac): Using unpackb since we would need to manage
        #   a buffer for Unpacker() which wouldn't gain us much.
        return self._unpackb(payload, raw=False)


class UnpackMethod(Protocol):
    def __call__(self, data: bytes, raw: bool = ...) -> Any: ...



class CustomHeadersMiddleware:
    def __init__(self, app, title_case=True, custom_capitalization=None):
        self._app = app
        self._title_case = title_case
        self._capitalization = custom_capitalization or {}

    def __call__(self, environ, start_response):
        def start_response_wrapper(status, response_headers, exc_info=None):
            if self._title_case:
                headers = [
                    (self._capitalization.get(name, name.title()), value)
                    for name, value in response_headers
                ]
            else:
                headers = [
                    (self._capitalization.get(name, name), value)
                    for name, value in response_headers
                ]
            start_response(status, headers, exc_info)

        return self._app(environ, start_response_wrapper)



import asyncio
from contextlib import contextmanager
import hashlib
import os
import platform
import random
import signal
import subprocess
import sys
import time

import pytest

try:
    import httpx
except ImportError:
    httpx = None  # type: ignore

try:
    import requests
    import requests.exceptions
except ImportError:
    requests = None  # type: ignore

try:
    import websockets
    import websockets.asyncio.client
    import websockets.exceptions
except ImportError:
    websockets = None  # type: ignore


from falcon import testing

from . import _asgi_test_app

_MODULE_DIR = os.path.abspath(os.path.dirname(__file__))

_PYPY = platform.python_implementation() == 'PyPy'
_WIN32 = sys.platform.startswith('win')

_SERVER_HOST = '127.0.0.1'
_SIZE_1_KB = 1024
_SIZE_1_MB = _SIZE_1_KB**2
# NOTE(vytas): Windows specific: {Application Exit by CTRL+C}.
#   The application terminated as a result of a CTRL+C.
_STATUS_CONTROL_C_EXIT = 0xC000013A

_REQUEST_TIMEOUT = 10


@pytest.mark.skipif(
    requests is None, reason='requests module is required for this test'
)
class TestASGIServer:
    def test_get(self, server_base_url):
        resp = requests.get(server_base_url, timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 200
        assert resp.text == '127.0.0.1'

    def test_put(self, server_base_url):
        body = '{}'
        resp = requests.put(server_base_url, data=body, timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 200
        assert resp.text == '{}'

    def test_head_405(self, server_base_url):
        body = '{}'
        resp = requests.head(server_base_url, data=body, timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 405

    def test_post_multipart_form(self, server_base_url):
        size = random.randint(16 * _SIZE_1_MB, 32 * _SIZE_1_MB)
        data = os.urandom(size)
        digest = hashlib.sha1(data).hexdigest()
        files = {
            'random': ('random.dat', data),
            'message': ('hello.txt', b'Hello, World!\n'),
        }

        resp = requests.post(
            server_base_url + 'forms', files=files, timeout=_REQUEST_TIMEOUT
        )
        assert resp.status_code == 200
        assert resp.json() == {
            'message': {
                'filename': 'hello.txt',
                'sha1': '60fde9c2310b0d4cad4dab8d126b04387efba289',
            },
            'random': {
                'filename': 'random.dat',
                'sha1': digest,
            },
        }

    def test_post_multiple(self, server_base_url):
        body = testing.rand_string(_SIZE_1_KB // 2, _SIZE_1_KB)
        resp = requests.post(server_base_url, data=body, timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 200
        assert resp.text == body
        assert resp.headers['X-Counter'] == '0'

        time.sleep(1)

        resp = requests.post(server_base_url, data=body, timeout=_REQUEST_TIMEOUT)
        assert resp.headers['X-Counter'] == '2002'

    def test_post_invalid_content_length(self, server_base_url):
        headers = {'Content-Length': 'invalid'}

        try:
            resp = requests.post(
                server_base_url, headers=headers, timeout=_REQUEST_TIMEOUT
            )

            # Daphne responds with a 400
            assert resp.status_code == 400

        except requests.ConnectionError:
            # NOTE(kgriffs): Uvicorn will kill the request so it does not
            #   even get to our app; the app logic is tested on the WSGI
            #   side. We leave this here in case something changes in
            #   the way uvicorn handles it or something and we want to
            #   get a heads-up if the request is no longer blocked.
            pass

    def test_post_read_bounded_stream(self, server_base_url):
        body = testing.rand_string(_SIZE_1_KB // 2, _SIZE_1_KB)
        resp = requests.post(
            server_base_url + 'bucket', data=body, timeout=_REQUEST_TIMEOUT
        )
        assert resp.status_code == 200
        assert resp.text == body

    def test_post_read_bounded_stream_large(self, server_base_url):
        """Test that we can correctly read large bodies chunked server-side.

        ASGI servers typically employ some type of flow control to stream
        large request bodies to the app. This occurs regardless of whether
        "chunked" Transfer-Encoding is employed by the client.
        """

        # NOTE(kgriffs): One would hope that flow control is effective enough
        #   to at least prevent bursting over 1 MB.
        size_mb = 5

        body = os.urandom(_SIZE_1_MB * size_mb)
        resp = requests.put(
            server_base_url + 'bucket/drops', data=body, timeout=_REQUEST_TIMEOUT
        )
        assert resp.status_code == 200
        assert resp.json().get('drops') > size_mb
        assert resp.json().get('sha1') == hashlib.sha1(body).hexdigest()

    def test_post_read_bounded_stream_no_body(self, server_base_url):
        resp = requests.post(server_base_url + 'bucket', timeout=_REQUEST_TIMEOUT)
        assert not resp.text

    def test_sse(self, server_base_url):
        resp = requests.get(server_base_url + 'events', timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 200

        events = resp.text.split('\n\n')
        assert len(events) > 2
        for e in events[:-1]:
            assert e == 'data: hello world'

        assert not events[-1]

    def test_sse_client_disconnects_early(self, server_base_url):
        """Test that when the client connection is lost, the server task does not hang.

        In the case of SSE, Falcon should detect when the client connection is
        lost and immediately bail out. Currently this is observable by watching
        the output of the uvicorn and daphne server processes. Also, the
        _run_server_isolated() method will fail the test if the server process
        takes too long to shut down.
        """
        with pytest.raises(requests.exceptions.ConnectionError):
            requests.get(
                server_base_url + 'events',
                timeout=(_asgi_test_app.SSE_TEST_MAX_DELAY_SEC / 2),
            )

    @pytest.mark.skipif(httpx is None, reason='httpx is required for this test')
    async def test_stream_chunked_request(self, server_base_url):
        """Regression test for https://github.com/falconry/falcon/issues/2024"""

        async def emitter():
            for _ in range(64):
                yield b'123456789ABCDEF\n'

        async with httpx.AsyncClient() as client:
            resp = await client.put(
                server_base_url + 'bucket/drops',
                content=emitter(),
                timeout=_REQUEST_TIMEOUT,
            )
            resp.raise_for_status()
            assert resp.json().get('drops') >= 1


@pytest.mark.skipif(
    requests is None, reason='requests module is required for this test'
)
@pytest.mark.skipif(
    websockets is None, reason='websockets is required for this test class'
)
class TestWebSocket:
    @pytest.mark.parametrize('explicit_close', [True, False])
    @pytest.mark.parametrize('close_code', [None, 4321])
    @pytest.mark.parametrize('max_receive_queue', [0, 4, 17])
    async def test_hello(
        self,
        explicit_close,
        close_code,
        max_receive_queue,
        server_base_url,
        server_url_events_ws,
    ):
        resp = requests.patch(
            server_base_url + 'wsoptions', json={'max_receive_queue': max_receive_queue}
        )
        resp.raise_for_status()

        echo_expected = 'Check 1 - \U0001f600'

        extra_headers = {'X-Command': 'recv'}

        if explicit_close:
            extra_headers['X-Close'] = 'True'

        if close_code:
            extra_headers['X-Close-Code'] = str(close_code)

        async with websockets.asyncio.client.connect(
            server_url_events_ws,
            additional_headers=extra_headers,
        ) as ws:
            got_message = False

            while True:
                try:
                    # TODO: Why is this failing to decode on the other side?
                    #   (raises an error)
                    # TODO: Why does this cause Daphne to hang?
                    await ws.send(f'{{"command": "echo", "echo": "{echo_expected}"}}')

                    message_text = await ws.recv()
                    message_echo = await ws.recv()
                    message_binary = await ws.recv()
                except websockets.exceptions.ConnectionClosed as ex:
                    if explicit_close and close_code:
                        assert ex.rcvd.code == close_code
                    else:
                        assert ex.rcvd.code == 1000

                    break

                got_message = True
                assert message_text == 'hello world'
                assert message_echo == echo_expected
                assert message_binary == b'hello\x00world'

            assert got_message

    @pytest.mark.parametrize('explicit_close', [True, False])
    @pytest.mark.parametrize('close_code', [None, 4040])
    async def test_rejected(self, explicit_close, close_code, server_url_events_ws):
        extra_headers = {'X-Accept': 'reject'}
        if explicit_close:
            extra_headers['X-Close'] = 'True'

        if close_code:
            extra_headers['X-Close-Code'] = str(close_code)

        with pytest.raises(websockets.exceptions.InvalidStatus) as exc_info:
            async with websockets.asyncio.client.connect(
                server_url_events_ws, additional_headers=extra_headers
            ):
                pass

        assert exc_info.value.response.status_code == 403

    async def test_missing_responder(self, server_url_events_ws):
        server_url_events_ws += '/404'

        with pytest.raises(websockets.exceptions.InvalidStatus) as exc_info:
            async with websockets.asyncio.client.connect(server_url_events_ws):
                pass

        assert exc_info.value.response.status_code == 403

    @pytest.mark.parametrize(
        'subprotocol, expected',
        [
            ('*', 'amqp'),
            ('wamp', 'wamp'),
        ],
    )
    async def test_select_subprotocol_known(
        self, subprotocol, expected, server_url_events_ws
    ):
        extra_headers = {'X-Subprotocol': subprotocol}
        async with websockets.asyncio.client.connect(
            server_url_events_ws,
            additional_headers=extra_headers,
            subprotocols=['amqp', 'wamp'],
        ) as ws:
            assert ws.subprotocol == expected

    async def test_select_subprotocol_unknown(self, server_url_events_ws):
        extra_headers = {'X-Subprotocol': 'xmpp'}

        try:
            async with websockets.asyncio.client.connect(
                server_url_events_ws,
                additional_headers=extra_headers,
                subprotocols=['amqp', 'wamp'],
            ):
                pass

            # NOTE(kgriffs): Taking the approach of asserting inside
            #   except clauses is a little bit cleaner in this case vs.
            #   multiple pytest.raises(), so we fail the test if no
            #   error is raised as expected.
            pytest.fail('no error raised')

        # Uvicorn
        except websockets.exceptions.NegotiationError as ex:
            assert 'unsupported subprotocol: xmpp' in str(ex)

        # Daphne, Hypercorn with websockets<14.2
        except EOFError:
            pass

        # Daphne, Hypercorn with websockets>=14.2
        except websockets.exceptions.InvalidMessage as ex:
            assert isinstance(ex.__cause__, EOFError)

    # NOTE(kgriffs): When executing this test under pytest with the -s
    #   argument, one should be able to see the message
    #   "on_websocket:WebSocketDisconnected" printed to the console. I have
    #   tried to capture this output and check it in the test below,
    #   but the usual ways of capturing stdout/stderr with pytest do
    #   not work.
    async def test_disconnecting_client_early(self, server_url_events_ws):
        ws = await websockets.asyncio.client.connect(
            server_url_events_ws, additional_headers={'X-Close': 'True'}
        )
        await asyncio.sleep(0.2)

        message_text = await ws.recv()
        assert message_text == 'hello world'

        message_binary = await ws.recv()
        assert message_binary == b'hello\x00world'

        await ws.close()
        print('closed')

        # NOTE(kgriffs): Let the app continue to attempt to send us
        #   messages after the close.
        await asyncio.sleep(1)

    async def test_send_before_accept(self, server_url_events_ws):
        extra_headers = {'x-accept': 'skip'}

        async with websockets.asyncio.client.connect(
            server_url_events_ws, additional_headers=extra_headers
        ) as ws:
            message = await ws.recv()
            assert message == 'OperationNotAllowed'

    async def test_recv_before_accept(self, server_url_events_ws):
        extra_headers = {'x-accept': 'skip', 'x-command': 'recv'}

        async with websockets.asyncio.client.connect(
            server_url_events_ws, additional_headers=extra_headers
        ) as ws:
            message = await ws.recv()
            assert message == 'OperationNotAllowed'

    async def test_invalid_close_code(self, server_url_events_ws):
        extra_headers = {'x-close': 'True', 'x-close-code': 42}

        async with websockets.asyncio.client.connect(
            server_url_events_ws, additional_headers=extra_headers
        ) as ws:
            start = time.time()

            while True:
                message = await asyncio.wait_for(ws.recv(), timeout=1)
                if message == 'ValueError':
                    break

                elapsed = time.time() - start
                assert elapsed < 2

    async def test_close_code_on_unhandled_error(self, server_url_events_ws):
        extra_headers = {'x-raise-error': 'generic'}

        async with websockets.asyncio.client.connect(
            server_url_events_ws, additional_headers=extra_headers
        ) as ws:
            await ws.wait_closed()

        assert ws.protocol.close_code in {3011, 1011}

    async def test_close_code_on_unhandled_http_error(self, server_url_events_ws):
        extra_headers = {'x-raise-error': 'http'}

        async with websockets.asyncio.client.connect(
            server_url_events_ws, additional_headers=extra_headers
        ) as ws:
            await ws.wait_closed()

        assert ws.protocol.close_code == 3400

    @pytest.mark.parametrize('mismatch', ['send', 'recv'])
    @pytest.mark.parametrize('mismatch_type', ['text', 'data'])
    async def test_type_mismatch(self, mismatch, mismatch_type, server_url_events_ws):
        extra_headers = {
            'X-Mismatch': mismatch,
            'X-Mismatch-Type': mismatch_type,
        }

        async with websockets.asyncio.client.connect(
            server_url_events_ws, additional_headers=extra_headers
        ) as ws:
            if mismatch == 'recv':
                if mismatch_type == 'text':
                    await ws.send(b'hello')
                else:
                    await ws.send('hello')

            await ws.wait_closed()

        assert ws.protocol.close_code in {3011, 1011}

    async def test_passing_path_params(self, server_base_url_ws):
        expected_feed_id = '1ee7'
        url = f'{server_base_url_ws}feeds/{expected_feed_id}'

        async with websockets.asyncio.client.connect(url) as ws:
            feed_id = await ws.recv()
            assert feed_id == expected_feed_id


@contextmanager
def _run_server_isolated(process_factory, host, port):
    # NOTE(kgriffs): We have to use subprocess because uvicorn has a tendency
    #   to corrupt our asyncio state and cause intermittent hangs in the test
    #   suite.
    print('\n[Starting server process...]')
    server = process_factory(host, port)

    yield server

    if _WIN32:
        # NOTE(kgriffs): Calling server.terminate() is equivalent to
        #   server.kill() on Windows. We don't want to do the this;
        #   forcefully killing a proc causes the CI job to fail,
        #   regardless of the tox/pytest exit code. """
        #
        #   Instead, we send CTRL+C. This does require that the handler be
        #   enabled via SetConsoleCtrlHandler() in _uvicorn_factory()
        #   below. Alternatively, we could send CTRL+BREAK and allow
        #   the process exit code to be 3221225786.
        #
        import signal

        print('\n[Sending CTRL+C (SIGINT) to server process...]')
        server.send_signal(signal.CTRL_C_EVENT)
        try:
            server.wait(timeout=10)
        except KeyboardInterrupt:
            pass
        except subprocess.TimeoutExpired:
            print('\n[Killing stubborn server process...]')

            server.kill()
            server.communicate()

            pytest.fail(
                'Server process did not exit in a timely manner and had to be killed.'
            )
    else:
        print('\n[Sending SIGTERM to server process...]')
        server.terminate()

        try:
            server.communicate(timeout=10)
        except subprocess.TimeoutExpired:
            print('\n[Killing stubborn server process...]')

            server.kill()
            server.communicate()

            pytest.fail(
                'Server process did not exit in a timely manner and had to be killed.'
            )


def _uvicorn_factory(host, port):
    if _WIN32:
        script = f"""
import uvicorn
import ctypes
ctypes.windll.kernel32.SetConsoleCtrlHandler(None, 0)
uvicorn.run('_asgi_test_app:application', host='{host}', port={port})
"""
        return subprocess.Popen(
            (sys.executable, '-c', script),
            cwd=_MODULE_DIR,
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
        )

    # NOTE(vytas): uvicorn+uvloop is not (well) supported on PyPy at the time
    #   of writing.
    loop_options = ('--http', 'h11', '--loop', 'asyncio') if _PYPY else ()
    options = (
        '--host',
        host,
        '--port',
        str(port),
        '--interface',
        'asgi3',
        '_asgi_test_app:application',
    )

    return subprocess.Popen(
        (
            sys.executable,
            '-m',
            'uvicorn',
        )
        + loop_options
        + options,
        cwd=_MODULE_DIR,
    )


def _daphne_factory(host, port):
    return subprocess.Popen(
        (
            sys.executable,
            '-m',
            'daphne',
            '--bind',
            host,
            '--port',
            str(port),
            '--verbosity',
            '2',
            '--access-log',
            '-',
            '_asgi_test_app:application',
        ),
        cwd=_MODULE_DIR,
    )


def _hypercorn_factory(host, port):
    if _WIN32:
        script = f"""
from hypercorn.run import Config, run
import ctypes
ctypes.windll.kernel32.SetConsoleCtrlHandler(None, 0)
config = Config()
config.application_path = '_asgi_test_app:application'
config.bind = ['{host}:{port}']
config.accesslog = '-'
config.debug = True
run(config)
"""
        return subprocess.Popen(
            (sys.executable, '-c', script),
            cwd=_MODULE_DIR,
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
        )
    return subprocess.Popen(
        (
            sys.executable,
            '-m',
            'hypercorn',
            '--bind',
            f'{host}:{port}',
            '--access-logfile',
            '-',
            '--debug',
            '_asgi_test_app:application',
        ),
        cwd=_MODULE_DIR,
    )


def _can_run(factory):
    if _WIN32 and factory == _daphne_factory:
        pytest.skip('daphne does not support windows')

    if factory == _daphne_factory:
        try:
            import daphne  # noqa
        except Exception:
            pytest.skip('daphne not installed')
    elif factory == _hypercorn_factory:
        try:
            import hypercorn  # noqa
        except Exception:
            pytest.skip('hypercorn not installed')
    elif factory == _uvicorn_factory:
        try:
            import uvicorn  # noqa
        except Exception:
            pytest.skip('uvicorn not installed')


@pytest.fixture(params=[_uvicorn_factory, _daphne_factory, _hypercorn_factory])
def server_base_url(request):
    process_factory = request.param
    _can_run(process_factory)

    for i in range(3):
        server_port = testing.get_unused_port()
        base_url = 'http://{}:{}/'.format(_SERVER_HOST, server_port)

        with _run_server_isolated(process_factory, _SERVER_HOST, server_port) as server:
            # NOTE(kgriffs): Let the server start up. Give up after 5 seconds.
            start_ts = time.time()
            while (time.time() - start_ts) < 5:
                try:
                    requests.get(base_url, timeout=0.2)
                except (
                    requests.exceptions.Timeout,
                    requests.exceptions.ConnectionError,
                ):
                    time.sleep(0.2)
                else:
                    break
            else:
                if server.poll() is None:
                    pytest.fail('Server is not responding to requests')
                else:
                    # NOTE(kgriffs): The server did not start up; probably due to
                    #   the port being in use. We could check the output but
                    #   capsys fixture may not have buffered the error output
                    #   yet, so we just retry.
                    continue

            yield base_url

        # NOTE(vytas): Starting with 0.29.0, Uvicorn will propagate signal
        #   values into the return code (which is a good practice in Unix);
        #   see also https://github.com/encode/uvicorn/pull/1600
        assert server.returncode in (0, -signal.SIGTERM, _STATUS_CONTROL_C_EXIT)

        break

    else:
        pytest.fail('Could not start server')


@pytest.fixture
def server_base_url_ws(server_base_url):
    return server_base_url.replace('http', 'ws')


@pytest.fixture
def server_url_events_ws(server_base_url_ws):
    return server_base_url_ws + 'events'



from __future__ import annotations

from functools import wraps
from inspect import iscoroutinefunction
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING

import falcon

try:
    import jsonschema
except ImportError:  # pragma: nocover
    pass

if TYPE_CHECKING:
    import falcon as wsgi
    from falcon import asgi

Schema = Optional[Dict[str, Any]]
ResponderMethod = Callable[..., Any]


def validate(
    req_schema: Schema = None, resp_schema: Schema = None
) -> Callable[[ResponderMethod], ResponderMethod]:
    """Validate ``req.media`` using JSON Schema.

    This decorator provides standard JSON Schema validation via the
    ``jsonschema`` package available from PyPI. Semantic validation via
    the *format* keyword is enabled for the default checkers implemented
    by ``jsonschema.FormatChecker``.

    In the case of failed request media validation, an instance of
    :class:`~falcon.MediaValidationError` is raised by the decorator. By
    default, this error is rendered as a 400 (:class:`~falcon.HTTPBadRequest`)
    response with the ``title`` and ``description`` attributes explaining the
    validation failure, but this behavior can be modified by adding a
    custom error :func:`handler <falcon.App.add_error_handler>` for
    :class:`~falcon.MediaValidationError`.

    Note:
        The ``jsonschema`` package must be installed separately in order to use
        this decorator, as Falcon does not install it by default.

        See `json-schema.org <https://json-schema.org>`_ for more
        information on defining a compatible dictionary.

    Keyword Args:
        req_schema (dict): A dictionary that follows the JSON
            Schema specification. The request will be validated against this
            schema.
        resp_schema (dict): A dictionary that follows the JSON
            Schema specification. The response will be validated against this
            schema.

    Example:

        .. tab-set::

            .. tab-item:: WSGI

                .. code:: python

                    from falcon.media.validators import jsonschema

                    # -- snip --

                    @jsonschema.validate(my_post_schema)
                    def on_post(self, req, resp):

                    # -- snip --

            .. tab-item:: ASGI

                .. code:: python

                    from falcon.media.validators import jsonschema

                    # -- snip --

                    @jsonschema.validate(my_post_schema)
                    async def on_post(self, req, resp):

                    # -- snip --

    """

    def decorator(func: ResponderMethod) -> ResponderMethod:
        if iscoroutinefunction(func):
            return _validate_async(func, req_schema, resp_schema)

        return _validate(func, req_schema, resp_schema)

    return decorator


def _validate(
    func: ResponderMethod, req_schema: Schema = None, resp_schema: Schema = None
) -> ResponderMethod:
    @wraps(func)
    def wrapper(
        self: Any, req: wsgi.Request, resp: wsgi.Response, *args: Any, **kwargs: Any
    ) -> Any:
        if req_schema is not None:
            try:
                jsonschema.validate(
                    req.media, req_schema, format_checker=jsonschema.FormatChecker()
                )
            except jsonschema.ValidationError as ex:
                raise falcon.MediaValidationError(
                    title='Request data failed validation', description=ex.message
                ) from ex

        result = func(self, req, resp, *args, **kwargs)

        if resp_schema is not None:
            try:
                jsonschema.validate(
                    resp.media, resp_schema, format_checker=jsonschema.FormatChecker()
                )
            except jsonschema.ValidationError as ex:
                raise falcon.HTTPInternalServerError(
                    title='Response data failed validation'
                    # Do not return 'e.message' in the response to
                    # prevent info about possible internal response
                    # formatting bugs from leaking out to users.
                ) from ex

        return result

    return wrapper


def _validate_async(
    func: ResponderMethod, req_schema: Schema = None, resp_schema: Schema = None
) -> ResponderMethod:
    @wraps(func)
    async def wrapper(
        self: Any, req: asgi.Request, resp: asgi.Response, *args: Any, **kwargs: Any
    ) -> Any:
        if req_schema is not None:
            m = await req.get_media()

            try:
                jsonschema.validate(
                    m, req_schema, format_checker=jsonschema.FormatChecker()
                )
            except jsonschema.ValidationError as ex:
                raise falcon.MediaValidationError(
                    title='Request data failed validation', description=ex.message
                ) from ex

        result = await func(self, req, resp, *args, **kwargs)

        if resp_schema is not None:
            try:
                jsonschema.validate(
                    resp.media, resp_schema, format_checker=jsonschema.FormatChecker()
                )
            except jsonschema.ValidationError as ex:
                raise falcon.HTTPInternalServerError(
                    title='Response data failed validation'
                    # Do not return 'e.message' in the response to
                    # prevent info about possible internal response
                    # formatting bugs from leaking out to users.
                ) from ex

        return result

    return wrapper



import http

import pytest

import falcon
from falcon.http_status import HTTPStatus
import falcon.testing as testing


@pytest.fixture()
def client(asgi, util):
    app = util.create_app(asgi)
    app.add_route('/status', TestStatusResource())
    return testing.TestClient(app)


@pytest.fixture()
def hook_test_client(asgi, util):
    app = util.create_app(asgi)
    app.add_route('/status', TestHookResource())
    return testing.TestClient(app)


def before_hook(req, resp, resource, params):
    raise HTTPStatus(falcon.HTTP_200, headers={'X-Failed': 'False'}, text='Pass')


def after_hook(req, resp, resource):
    resp.status = falcon.HTTP_200
    resp.set_header('X-Failed', 'False')
    resp.text = 'Pass'


def noop_after_hook(req, resp, resource):
    pass


class TestStatusResource:
    @falcon.before(before_hook)
    def on_get(self, req, resp):
        resp.status = falcon.HTTP_500
        resp.set_header('X-Failed', 'True')
        resp.text = 'Fail'

    def on_post(self, req, resp):
        resp.status = falcon.HTTP_500
        resp.set_header('X-Failed', 'True')
        resp.text = 'Fail'

        raise HTTPStatus(falcon.HTTP_200, headers={'X-Failed': 'False'}, text='Pass')

    @falcon.after(after_hook)
    def on_put(self, req, resp):
        # NOTE(kgriffs): Test that passing a unicode status string
        # works just fine.
        resp.status = '500 Internal Server Error'
        resp.set_header('X-Failed', 'True')
        resp.text = 'Fail'

    def on_patch(self, req, resp):
        raise HTTPStatus(falcon.HTTP_200, text=None)

    @falcon.after(noop_after_hook)
    def on_delete(self, req, resp):
        raise HTTPStatus(201, headers={'X-Failed': 'False'}, text='Pass')


class TestHookResource:
    def on_get(self, req, resp):
        resp.status_code = 500
        resp.set_header('X-Failed', 'True')
        resp.text = 'Fail'

    def on_patch(self, req, resp):
        raise HTTPStatus(200, text=None)


class TestHTTPStatus:
    def test_raise_status_in_before_hook(self, client):
        """Make sure we get the 200 raised by before hook"""
        response = client.simulate_request(path='/status', method='GET')
        assert response.status == falcon.HTTP_200
        assert response.status_code == 200
        assert response.headers['x-failed'] == 'False'
        assert response.text == 'Pass'

    def test_raise_status_in_responder(self, client):
        """Make sure we get the 200 raised by responder"""
        response = client.simulate_request(path='/status', method='POST')
        assert response.status == falcon.HTTP_200
        assert response.status_code == 200
        assert response.headers['x-failed'] == 'False'
        assert response.text == 'Pass'

    def test_raise_status_runs_after_hooks(self, client):
        """Make sure after hooks still run"""
        response = client.simulate_request(path='/status', method='PUT')
        assert response.status == falcon.HTTP_200
        assert response.status_code == 200
        assert response.headers['x-failed'] == 'False'
        assert response.text == 'Pass'

    def test_raise_status_survives_after_hooks(self, client):
        """Make sure after hook doesn't overwrite our status"""
        response = client.simulate_request(path='/status', method='DELETE')
        assert response.status == falcon.HTTP_201
        assert response.status_code == 201
        assert response.headers['x-failed'] == 'False'
        assert response.text == 'Pass'

    def test_raise_status_empty_body(self, client):
        """Make sure passing None to body results in empty body"""
        response = client.simulate_request(path='/status', method='PATCH')
        assert response.text == ''


class TestHTTPStatusWithMiddleware:
    def test_raise_status_in_process_request(self, hook_test_client):
        """Make sure we can raise status from middleware process request"""
        client = hook_test_client

        class TestMiddleware:
            def process_request(self, req, resp):
                raise HTTPStatus(
                    falcon.HTTP_200, headers={'X-Failed': 'False'}, text='Pass'
                )

            # NOTE(kgriffs): Test the side-by-side support for dual WSGI and
            #   ASGI compatibility.
            async def process_request_async(self, req, resp):
                self.process_request(req, resp)

        client.app.add_middleware(TestMiddleware())

        response = client.simulate_request(path='/status', method='GET')
        assert response.status_code == 200
        assert response.headers['x-failed'] == 'False'
        assert response.text == 'Pass'

    def test_raise_status_in_process_resource(self, hook_test_client):
        """Make sure we can raise status from middleware process resource"""
        client = hook_test_client

        class TestMiddleware:
            def process_resource(self, req, resp, resource, params):
                raise HTTPStatus(
                    falcon.HTTP_200, headers={'X-Failed': 'False'}, text='Pass'
                )

            async def process_resource_async(self, *args):
                self.process_resource(*args)

        # NOTE(kgriffs): Pass a list to test that add_middleware can handle it
        client.app.add_middleware([TestMiddleware()])

        response = client.simulate_request(path='/status', method='GET')
        assert response.status == falcon.HTTP_200
        assert response.headers['x-failed'] == 'False'
        assert response.text == 'Pass'

    def test_raise_status_runs_process_response(self, hook_test_client):
        """Make sure process_response still runs"""
        client = hook_test_client

        class TestMiddleware:
            def process_response(self, req, resp, resource, req_succeeded):
                resp.status = falcon.HTTP_200
                resp.set_header('X-Failed', 'False')
                resp.text = 'Pass'

            async def process_response_async(self, *args):
                self.process_response(*args)

        # NOTE(kgriffs): Pass a generic iterable to test that add_middleware
        #   can handle it.
        client.app.add_middleware(iter([TestMiddleware()]))

        response = client.simulate_request(path='/status', method='GET')
        assert response.status == falcon.HTTP_200
        assert response.headers['x-failed'] == 'False'
        assert response.text == 'Pass'


class NoBodyResource:
    def on_get(self, req, res):
        res.data = b'foo'
        http_status = HTTPStatus(745)
        assert http_status.status_code == 745
        raise http_status

    def on_post(self, req, res):
        res.media = {'a': 1}
        http_status = HTTPStatus(falcon.HTTP_725)
        assert http_status.status_code == 725
        raise http_status

    def on_put(self, req, res):
        res.text = 'foo'
        raise HTTPStatus(falcon.HTTP_719)


@pytest.fixture()
def body_client(asgi, util):
    app = util.create_app(asgi)
    app.add_route('/status', NoBodyResource())
    return testing.TestClient(app)


class TestNoBodyWithStatus:
    def test_data_is_set(self, body_client):
        res = body_client.simulate_get('/status')
        assert res.status == falcon.HTTP_745
        assert res.status_code == 745
        assert res.content == b''

    def test_media_is_set(self, body_client):
        res = body_client.simulate_post('/status')
        assert res.status == falcon.HTTP_725
        assert res.status_code == 725
        assert res.content == b''

    def test_body_is_set(self, body_client):
        res = body_client.simulate_put('/status')
        assert res.status == falcon.HTTP_719
        assert res.status_code == 719
        assert res.content == b''


@pytest.fixture()
def custom_status_client(asgi, util):
    def client(status):
        class Resource:
            def on_get(self, req, resp):
                resp.content_type = falcon.MEDIA_TEXT
                resp.data = b'Hello, World!'
                resp.status = status

        app = util.create_app(asgi)
        app.add_route('/status', Resource())
        return testing.TestClient(app)

    return client


@pytest.mark.parametrize(
    'status,expected_code',
    [
        (http.HTTPStatus(200), 200),
        (http.HTTPStatus(202), 202),
        (http.HTTPStatus(403), 403),
        (http.HTTPStatus(500), 500),
        (http.HTTPStatus.OK, 200),
        (http.HTTPStatus.USE_PROXY, 305),
        (http.HTTPStatus.NOT_FOUND, 404),
        (http.HTTPStatus.NOT_IMPLEMENTED, 501),
        (200, 200),
        (307, 307),
        (500, 500),
        (702, 702),
        (b'200 OK', 200),
        (b'702 Emacs', 702),
    ],
)
def test_non_string_status(custom_status_client, status, expected_code):
    client = custom_status_client(status)
    resp = client.simulate_get('/status')
    assert resp.text == 'Hello, World!'
    assert resp.status_code == expected_code


def test_deprecated_body():
    with pytest.raises(TypeError) as type_error:
        sts = HTTPStatus(falcon.HTTP_701, body='foo')

    assert 'unexpected keyword argument' in str(type_error.value)

    sts = HTTPStatus(falcon.HTTP_701, text='foo')
    assert sts.text == 'foo'



import itertools

import pytest

from falcon import errors
from falcon.util import mediatypes


@pytest.mark.parametrize(
    'value,expected',
    [
        ('', ('', {})),
        ('strange', ('strange', {})),
        ('text/plain', ('text/plain', {})),
        ('text/plain ', ('text/plain', {})),
        (' text/plain', ('text/plain', {})),
        (' text/plain ', ('text/plain', {})),
        ('   text/plain   ', ('text/plain', {})),
        (
            'falcon/peregrine;  key1; key2=value; key3',
            ('falcon/peregrine', {'key2': 'value'}),
        ),
        (
            'audio/pcm;rate=48000;encoding=float;bits=32',
            ('audio/pcm', {'bits': '32', 'encoding': 'float', 'rate': '48000'}),
        ),
        (
            'falcon/*; genus=falco; family=falconidae; class=aves; ',
            ('falcon/*', {'class': 'aves', 'family': 'falconidae', 'genus': 'falco'}),
        ),
        ('"falcon/peregrine" ; key="value"', ('"falcon/peregrine"', {'key': 'value'})),
        ('falcon/peregrine; empty=""', ('falcon/peregrine', {'empty': ''})),
        ('falcon/peregrine; quote="', ('falcon/peregrine', {'quote': '"'})),
        ('text/plain; charset=utf-8', ('text/plain', {'charset': 'utf-8'})),
        ('stuff/strange; missing-value; missing-another', ('stuff/strange', {})),
        ('stuff/strange; missing-value\\missing-another', ('stuff/strange', {})),
        (
            'application/falcon; P1 = "key; value"; P2="\\""',
            ('application/falcon', {'p1': 'key; value', 'p2': '"'}),
        ),
    ],
)
def test_parse_header(value, expected):
    assert mediatypes.parse_header(value) == expected


def test_media_type_private_cls():
    mt1 = mediatypes._MediaType.parse('image/png')
    assert mt1.main_type == 'image'
    assert mt1.subtype == 'png'
    assert mt1.params == {}

    mt2 = mediatypes._MediaType.parse('text/plain; charset=latin-1')
    assert mt2.main_type == 'text'
    assert mt2.subtype == 'plain'
    assert mt2.params == {'charset': 'latin-1'}


def test_media_range_private_cls():
    mr1 = mediatypes._MediaRange.parse('image/png')
    assert mr1.main_type == 'image'
    assert mr1.subtype == 'png'
    assert mr1.quality == 1.0
    assert mr1.params == {}

    mr2 = mediatypes._MediaRange.parse('text/plain; charset=latin-1; Q=0.9')
    assert mr2.main_type == 'text'
    assert mr2.subtype == 'plain'
    assert pytest.approx(mr2.quality) == 0.9
    assert mr2.params == {'charset': 'latin-1'}

    mr3 = mediatypes._MediaRange.parse('*; q=0.7')
    assert mr3.main_type == '*'
    assert mr3.subtype == '*'
    assert pytest.approx(mr3.quality) == 0.7
    assert mr3.params == {}


_RFC_7231_EXAMPLE_ACCEPT = (
    'text/*;q=0.3, text/html;q=0.7, text/html;level=1, '
    'text/html;level=2;q=0.4, */*;q=0.5'
)
_RFC_7231_EXAMPLE_VALUES = [
    ('text/html;level=1', 1.0),
    ('text/html', 0.7),
    ('text/plain', 0.3),
    ('image/jpeg', 0.5),
    ('text/html;level=2', 0.4),
    ('text/html;level=3', 0.7),
]

_RFC_9110_EXAMPLE_ACCEPT = (
    'text/*;q=0.3, text/plain;q=0.7, text/plain;format=flowed, '
    'text/plain;format=fixed;q=0.4, */*;q=0.5'
)
# NOTE(vytas): Including the errata https://www.rfc-editor.org/errata/eid7138.
_RFC_9110_EXAMPLE_VALUES = [
    ('text/plain;format=flowed', 1.0),
    ('text/plain', 0.7),
    ('text/html', 0.3),
    ('image/jpeg', 0.5),
    ('text/plain;format=fixed', 0.4),
    ('text/html;level=3', 0.3),
]

_RFC_EXAMPLES = list(
    itertools.chain.from_iterable(
        ((accept,) + media_type_quality for media_type_quality in example_values)
        for accept, example_values in (
            (_RFC_7231_EXAMPLE_ACCEPT, _RFC_7231_EXAMPLE_VALUES),
            (_RFC_9110_EXAMPLE_ACCEPT, _RFC_9110_EXAMPLE_VALUES),
        )
    )
)

_RFC_EXAMPLE_IDS = list(
    itertools.chain.from_iterable(
        (
            f'{rfc}-{media_type}-{quality_value}'
            for media_type, quality_value in example_values
        )
        for rfc, example_values in (
            ('RFC-7231', _RFC_7231_EXAMPLE_VALUES),
            ('RFC-9110', _RFC_7231_EXAMPLE_VALUES),
        )
    )
)


@pytest.mark.parametrize(
    'accept,media_type,quality_value', _RFC_EXAMPLES, ids=_RFC_EXAMPLE_IDS
)
def test_quality_rfc_examples(accept, media_type, quality_value):
    assert pytest.approx(mediatypes.quality(media_type, accept)) == quality_value


@pytest.mark.parametrize(
    'accept,media_type,quality_value',
    [
        (
            'application/*, */wildcard; q=0.7, */*; q=0.25',
            'test/wildcard; expect=pass',
            0.7,
        ),
        (
            'application/*, */wildcard; q=0.7, */*; q=0.25',
            'application/wildcard; expect=pass',
            1.0,
        ),
        (
            'application/*, */wildcard; q=0.7, */*; q=0.25',
            'test/something; expect=pass',
            0.25,
        ),
        (
            'text/x-python, text/*; q=0.33, text/plain; format=fixed',
            'text/plain; format=flowed',
            0.33,
        ),
        (
            # NOTE(vytas): Same as one of the RFC 7231 examples, just with some
            #   media ranges reordered. python-mimeparse fails to yield the
            #   correct result in this specific case.
            'text/*;q=0.3, text/html;level=1, text/html;q=0.7, '
            'text/html;level=2;q=0.4, */*;q=0.5',
            'text/html; level=3',
            0.7,
        ),
    ],
)
def test_quality(accept, media_type, quality_value):
    assert pytest.approx(mediatypes.quality(media_type, accept)) == quality_value


@pytest.mark.parametrize(
    'accept,media_type',
    [
        (
            'foo/bar, test/app; q=0.2, test/app; p=1; q=0.9, test/app;p=1;r=2',
            'test/app',
        ),
        ('test/app; q=0.1, test/app; p=1; q=0.2, test/app;p=1;r=2', 'test/app; p=1'),
        (
            '*/app; q=0.1, simple/app; test=true; q=0.2, simple/app; color=blue',
            'simple/app; test=true',
        ),
    ],
)
def test_quality_prefer_exact_match(accept, media_type):
    assert pytest.approx(mediatypes.quality(media_type, accept)) == 0.2


_QUALITY_NONE_MATCHES_EXAMPLES = [
    ('application/json', 'application/yaml'),
    ('audio/*; q=0.2, audio/basic', 'video/mp3'),
    (
        'falcon/peregrine; speed=high; unladen=true',
        'falcon/peregrine; speed=average',
    ),
    ('text/html, text/plain', 'text/x-python'),
    ('*/json; q=0.2, application/json', 'application/msgpack'),
    (
        'text/x-python, image/*; q=0.33, text/plain; format=fixed',
        'text/plain; format=flowed',
    ),
]


@pytest.mark.parametrize('accept,media_type', _QUALITY_NONE_MATCHES_EXAMPLES)
def test_quality_none_matches(accept, media_type):
    assert mediatypes.quality(media_type, accept) == 0.0


@pytest.mark.parametrize(
    'media_types,accept,expected',
    [
        (['application/json'], 'application/json', 'application/json'),
        (['application/json'], 'application/json; charset=utf-8', 'application/json'),
        (
            ['application/json', 'application/yaml'],
            'application/json, */*; q=0.2',
            'application/json',
        ),
    ],
)
def test_best_match(media_types, accept, expected):
    assert mediatypes.best_match(media_types, accept) == expected


_BEST_MATCH_NONE_MATCHES_EXAMPLES = [
    ([_mt], _acc) for _mt, _acc in _QUALITY_NONE_MATCHES_EXAMPLES
] + [
    (['application/json', 'application/yaml'], 'application/xml, text/*; q=0.7'),
    (
        ['text/plain', 'falcon/peregrine; load=unladen'],
        'falcon/peregrine; load=heavy',
    ),
]


@pytest.mark.parametrize('media_types,accept', _BEST_MATCH_NONE_MATCHES_EXAMPLES)
def test_best_match_none_matches(media_types, accept):
    assert mediatypes.best_match(media_types, accept) == ''


@pytest.mark.parametrize('media_type', ['', 'word document', 'text'])
def test_invalid_media_type(media_type):
    with pytest.raises(errors.InvalidMediaType):
        mediatypes.quality(media_type, '*/*')


def _generate_strings(items):
    yield from items


@pytest.mark.parametrize(
    'media_range',
    [
        '',
        'word document',
        'text',
        'text/plain; q=high',
        '*/*; q=inf',
        '*/*; q=-inf',
        '*/*; q=nan',
        'application/very-important; q=1337.0',
    ],
)
def test_invalid_media_range(media_range):
    with pytest.raises(errors.InvalidMediaRange):
        mediatypes.quality('falcon/peregrine', media_range)

    with pytest.raises(errors.InvalidMediaRange):
        mediatypes.best_match(_generate_strings(['falcon/peregrine']), media_range)


@pytest.mark.parametrize(
    'accept',
    ['*/*', 'application/xml, text/*; q=0.7, */*; q=0.3'],
)
@pytest.mark.parametrize('media_types', [(), [], {}, _generate_strings(())])
def test_empty_media_types(accept, media_types):
    assert mediatypes.best_match(media_types, accept) == ''



import io
import os
from unittest import TestCase
from unittest.mock import call
from unittest.mock import MagicMock
from unittest.mock import mock_open
import uuid
from wsgiref.validate import InputWrapper

import pytest

import falcon
from falcon import testing

import look.app
import look.images


@pytest.fixture
def mock_store():
    return MagicMock()


@pytest.fixture
def client(mock_store):
    api = look.app.create_app(mock_store)
    return testing.TestClient(api)


def test_list_images(client, mock_store):
    images = ['first-file', 'second-file', 'third-file']
    image_docs = [{'href': '/images/' + image} for image in images]

    mock_store.list.return_value = images

    response = client.simulate_get('/images')

    result = response.json

    assert result['images'] == image_docs
    assert response.status == falcon.HTTP_OK


# With clever composition of fixtures, we can observe what happens with
# the mock injected into the image resource.
def test_post_image(client, mock_store):
    file_name = 'fake-image-name.xyz'

    # We need to know what ImageStore method will be used
    mock_store.save.return_value = file_name
    image_content_type = 'image/xyz'

    response = client.simulate_post(
        '/images', body=b'some-fake-bytes', headers={'content-type': image_content_type}
    )

    assert response.status == falcon.HTTP_CREATED
    assert response.headers['location'] == '/images/{}'.format(file_name)
    saver_call = mock_store.save.call_args

    # saver_call is a unittest.mock.call tuple. It's first element is a
    # tuple of positional arguments supplied when calling the mock.
    assert isinstance(saver_call[0][0], InputWrapper)
    assert saver_call[0][1] == image_content_type


def test_saving_image():
    # This still has some mocks, but they are more localized and do not
    # have to be monkey-patched into standard library modules (always a
    # risky business).
    mock_file_open = mock_open()

    fake_uuid = '123e4567-e89b-12d3-a456-426655440000'

    def mock_uuidgen():
        return fake_uuid

    fake_image_bytes = b'fake-image-bytes'
    fake_request_stream = io.BytesIO(fake_image_bytes)
    storage_path = 'fake-storage-path'
    store = look.images.ImageStore(
        storage_path, uuidgen=mock_uuidgen, fopen=mock_file_open
    )

    assert store.save(fake_request_stream, 'image/png') == fake_uuid + '.png'
    assert call().write(fake_image_bytes) in mock_file_open.mock_calls


def test_get_image(client, mock_store):
    file_bytes = b'fake-image-bytes'

    mock_store.open.return_value = ((file_bytes,), 17)

    response = client.simulate_get('/images/filename.png')

    assert response.status == falcon.HTTP_OK
    assert response.content == file_bytes


def test_opening_image():
    file_name = f'{uuid.uuid4()}.png'
    storage_path = '.'
    file_path = f'{storage_path}/{file_name}'
    fake_image_bytes = b'fake-image-bytes'
    with open(file_path, 'wb') as image_file:
        file_length = image_file.write(fake_image_bytes)

    store = look.images.ImageStore(storage_path)

    file_reader, content_length = store.open(file_name)
    assert content_length == file_length
    assert file_reader.read() == fake_image_bytes
    os.remove(file_path)

    with TestCase().assertRaises(IOError):
        store.open('wrong_file_name_format')


def test_listing_images():
    file_names = [f'{uuid.uuid4()}.png' for _ in range(2)]
    storage_path = '.'
    file_paths = [f'{storage_path}/{name}' for name in file_names]
    fake_images_bytes = [
        b'fake-image-bytes',  # 17
        b'fake-image-bytes-with-more-length',  # 34
    ]
    for i in range(2):
        with open(file_paths[i], 'wb') as image_file:
            image_file.write(fake_images_bytes[i])

    store = look.images.ImageStore(storage_path)
    assert store.list(10) == []
    assert store.list(20) == [file_names[0]]
    assert len(store.list(40)) == 2
    assert sorted(store.list(40)) == sorted(file_names)

    for file_path in file_paths:
        os.remove(file_path)



from __future__ import annotations

from typing import Optional

from falcon.constants import MEDIA_JSON
from falcon.media import BaseHandler
from falcon.media.json import _DEFAULT_JSON_HANDLER

__all__ = ('SSEvent',)


class SSEvent:
    """Represents a Server-Sent Event (SSE).

    Instances of this class can be yielded by an async generator in order to
    send a series of `Server-Sent Events`_ to the user agent.

    (See also: :attr:`falcon.asgi.Response.sse`)

    Keyword Args:
        data (bytes): Raw byte string to use as the ``data`` field for the
            event message. Takes precedence over both `text` and `json`.
        text (str): String to use for the ``data`` field in the message. Will
            be encoded as UTF-8 in the event. Takes precedence over `json`.
        json (object): JSON-serializable object to be converted to JSON and
            used as the ``data`` field in the event message.
        event (str): A string identifying the event type (AKA event name).
        event_id (str): The event ID that the User Agent should use for
            the `EventSource` object's last event ID value.
        retry (int): The reconnection time to use when attempting to send the
            event. This must be an integer, specifying the reconnection time
            in milliseconds.
        comment (str): Comment to include in the event message; this is
            normally ignored by the user agent, but is useful when composing
            a periodic "ping" message to keep the connection alive. Since this
            is a common use case, a default "ping" comment will be included
            in any event that would otherwise be blank (i.e., one that does
            not specify any fields when initializing the `SSEvent` instance.)

    .. _Server-Sent Events:
        https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events
    """

    __slots__ = [
        'data',
        'text',
        'json',
        'event',
        'event_id',
        'retry',
        'comment',
    ]

    data: Optional[bytes]
    """Raw byte string to use as the ``data`` field for the event message.
    Takes precedence over both `text` and `json`.
    """
    text: Optional[str]
    """String to use for the ``data`` field in the message.
    Will be encoded as UTF-8 in the event. Takes precedence over `json`.
    """
    json: object
    """JSON-serializable object to be converted to JSON and used as the ``data``
    field in the event message.
    """
    event: Optional[str]
    """A string identifying the event type (AKA event name)."""
    event_id: Optional[str]
    """The event ID that the User Agent should use for the `EventSource` object's
    last event ID value.
    """
    retry: Optional[int]
    """The reconnection time to use when attempting to send the event.

    This must be an integer, specifying the reconnection time in milliseconds.
    """
    comment: Optional[str]
    """Comment to include in the event message.

    This is normally ignored by the user agent, but is useful when composing a periodic
    "ping" message to keep the connection alive. Since this is a common use case, a
    default "ping" comment will be included in any event that would otherwise be blank
    (i.e., one that does not specify any of the fields when initializing the
    :class:`SSEvent` instance.)
    """

    def __init__(
        self,
        data: Optional[bytes] = None,
        text: Optional[str] = None,
        json: Optional[object] = None,
        event: Optional[str] = None,
        event_id: Optional[str] = None,
        retry: Optional[int] = None,
        comment: Optional[str] = None,
    ) -> None:
        # NOTE(kgriffs): Check up front since this makes it a lot easier
        #   to debug the source of the problem in the app vs. waiting for
        #   an error to be raised from the framework when it calls serialize()
        #   after the fact.

        if data is not None and not isinstance(data, bytes):
            raise TypeError('data must be a byte string')

        if text is not None and not isinstance(text, str):
            raise TypeError('text must be a string')

        if event is not None and not isinstance(event, str):
            raise TypeError('event name must be a string')

        if event_id is not None and not isinstance(event_id, str):
            raise TypeError('event_id must be a string')

        if comment is not None and not isinstance(comment, str):
            raise TypeError('comment must be a string')

        if retry is not None and not isinstance(retry, int):
            raise TypeError('retry must be an int')

        self.data = data
        self.text = text
        self.json = json
        self.event = event
        self.event_id = event_id
        self.retry = retry

        self.comment = comment

    def serialize(self, handler: Optional[BaseHandler] = None) -> bytes:
        """Serialize this event to string.

        Args:
            handler: Handler object that will be used to serialize the ``json``
                attribute to string. When not provided, a default handler using
                the builtin JSON library will be used (default ``None``).

        Returns:
            bytes: string representation of this event.
        """
        if self.comment is not None:
            block = f': {self.comment}\n'
        else:
            block = ''

        if self.event is not None:
            block += f'event: {self.event}\n'

        if self.event_id is not None:
            # NOTE(kgriffs): f-strings are a tiny bit faster than str().
            block += f'id: {self.event_id}\n'

        if self.retry is not None:
            block += f'retry: {self.retry}\n'

        if self.data is not None:
            # NOTE(kgriffs): While this decode() may seem unnecessary, it
            #   does provide a check to ensure it is valid UTF-8. I'm also
            #   assuming for the moment that most people will not use this
            #   attribute, but rather the text and json ones instead. If that
            #   is true, it makes sense to construct the entire string
            #   first, then encode it all in one go at the end.
            block += f'data: {self.data.decode()}\n'
        elif self.text is not None:
            block += f'data: {self.text}\n'
        elif self.json is not None:
            if handler is None:
                handler = _DEFAULT_JSON_HANDLER
            serialized = handler.serialize(self.json, MEDIA_JSON)
            block += 'data: '
            return block.encode() + serialized + b'\n\n'

        if not block:
            return b': ping\n\n'

        return (block + '\n').encode()



import asyncio
import json

import pytest

import falcon
from falcon import testing
from falcon.asgi import App
from falcon.asgi import SSEvent


def test_no_events():
    class Emitter:
        def __aiter__(self):
            return self

        async def __anext__(self):
            raise StopAsyncIteration

    class SomeResource:
        async def on_get(self, req, resp):
            self._called = True
            resp.sse = Emitter()

            # NOTE(vytas): Test explicitly referencing property since it is
            # normally optimized away by operating directly on the private
            # attribute in hot App.__call__ code paths.
            assert resp.sse is not None

    resource = SomeResource()

    app = App()
    app.add_route('/', resource)

    client = testing.TestClient(app)
    client.simulate_get()

    assert resource._called


def test_single_event():
    class SomeResource:
        async def on_get(self, req, resp):
            async def emitter():
                yield

            resp.sse = emitter()

        async def on_post(self, req, resp):
            async def emitter():
                yield SSEvent()

            resp.sse = emitter()

    resource = SomeResource()

    app = App()
    app.add_route('/', resource)

    client = testing.TestClient(app)

    result = client.simulate_get()
    assert result.text == ': ping\n\n'

    result = client.simulate_post()
    assert result.text == ': ping\n\n'


def test_multiple_events():
    expected_result_text = (
        'data: ketchup\n'
        '\n'
        'event: condiment\n'
        'data: mustard\n'
        '\n'
        'event: condiment\n'
        'id: 1234\n'
        'data: mayo\n'
        '\n'
        'event: topping\n'
        'id: 5678\n'
        'retry: 100\n'
        'data: onions\n'
        '\n'
        ': Serve with chips.\n'
        'retry: 100\n'
        'data: guacamole \u1f951\n'
        '\n'
        'retry: 100\n'
        'data: {"condiment": "salsa"}\n'
        '\n'
    )

    class SomeResource:
        async def on_get(self, req, resp):
            async def emitter():
                for event in [
                    SSEvent(data=b'ketchup'),
                    SSEvent(data=b'mustard', event='condiment'),
                    SSEvent(data=b'mayo', event='condiment', event_id='1234'),
                    SSEvent(
                        data=b'onions', event='topping', event_id='5678', retry=100
                    ),
                    SSEvent(
                        text='guacamole \u1f951', retry=100, comment='Serve with chips.'
                    ),
                    SSEvent(json={'condiment': 'salsa'}, retry=100),
                ]:
                    yield event
                    await asyncio.sleep(0.001)

            resp.sse = emitter()

    resource = SomeResource()

    app = App()
    app.add_route('/', resource)

    client = testing.TestClient(app)

    async def _test():
        async with client as conductor:
            # NOTE(kgriffs): Single-shot test will only allow the first
            #   one or two events since a client disconnect will be emitted
            #   into the app immediately.
            result = await conductor.simulate_get()
            assert expected_result_text.startswith(result.text)

            async with conductor.simulate_get_stream() as sr:
                event_count = 0

                result_text = ''

                while True:
                    chunk = (await sr.stream.read()).decode()

                    if not chunk:
                        continue

                    result_text += chunk
                    event_count += len(chunk.strip().split('\n\n'))

                    if 'salsa' in chunk:
                        break

                assert not (await sr.stream.read())

                assert event_count == 6
                assert result_text == expected_result_text

    falcon.async_to_sync(_test)


def test_multiple_events_early_disconnect():
    class SomeResource:
        async def on_get(self, req, resp):
            async def emitter():
                while True:
                    yield SSEvent(data=b'whassup')
                    await asyncio.sleep(0.01)

            resp.sse = emitter()

    resource = SomeResource()

    app = App()
    app.add_route('/', resource)

    async def _test():
        conductor = testing.ASGIConductor(app)
        result = await conductor.simulate_get()
        assert 'data: whassup' in result.text

        async with testing.ASGIConductor(app) as conductor:
            # NOTE(vytas): Using the get_stream() alias.
            async with conductor.get_stream() as sr:
                event_count = 0

                result_text = ''

                while event_count < 5:
                    chunk = (await sr.stream.read()).decode()
                    if not chunk:
                        continue

                    result_text += chunk
                    event_count += len(chunk.strip().split('\n\n'))

                assert result_text.startswith('data: whassup\n\n' * 5)
                assert event_count == 5

    falcon.async_to_sync(_test)


class TestSerializeJson:
    @pytest.fixture
    def client(self):
        class SomeResource:
            async def on_get(self, req, resp):
                async def emitter():
                    yield SSEvent(json={'foo': 'bar'})
                    yield SSEvent(json={'bar': 'baz'})

                resp.sse = emitter()

        resource = SomeResource()

        app = App()
        app.add_route('/', resource)

        client = testing.TestClient(app)
        return client

    def test_use_media_handler_dumps(self, client):
        h = client.app.resp_options.media_handlers[falcon.MEDIA_JSON]
        h._dumps = lambda x: json.dumps(x).upper()

        result = client.simulate_get()
        assert result.text == 'data: {"FOO": "BAR"}\n\ndata: {"BAR": "BAZ"}\n\n'

    def test_no_json_media_handler(self, client):
        for h in list(client.app.resp_options.media_handlers):
            if 'json' in h.casefold():
                client.app.resp_options.media_handlers.pop(h)

        result = client.simulate_get()
        assert result.text == ('data: {"foo": "bar"}\n\ndata: {"bar": "baz"}\n\n')


def test_invalid_event_values():
    with pytest.raises(TypeError):
        SSEvent(data='notbytes')

    with pytest.raises(TypeError):
        SSEvent(data=12345)

    with pytest.raises(TypeError):
        SSEvent(data=0)

    with pytest.raises(TypeError):
        SSEvent(text=b'notbytes')

    with pytest.raises(TypeError):
        SSEvent(text=23455)

    with pytest.raises(TypeError):
        SSEvent(text=0)

    with pytest.raises(TypeError):
        SSEvent(json=set()).serialize()

    with pytest.raises(TypeError):
        SSEvent(event=b'name')

    with pytest.raises(TypeError):
        SSEvent(event=1234)

    with pytest.raises(TypeError):
        SSEvent(event=0)

    with pytest.raises(TypeError):
        SSEvent(event_id=b'idbytes')

    with pytest.raises(TypeError):
        SSEvent(event_id=52085)

    with pytest.raises(TypeError):
        SSEvent(event_id=0)

    with pytest.raises(TypeError):
        SSEvent(retry='5808.25')

    with pytest.raises(TypeError):
        SSEvent(retry='')

    with pytest.raises(TypeError):
        SSEvent(retry=5808.25)

    with pytest.raises(TypeError):
        SSEvent(comment=b'somebytes')

    with pytest.raises(TypeError):
        SSEvent(comment=1234)

    with pytest.raises(TypeError):
        SSEvent(comment=0)


def test_non_iterable():
    class SomeResource:
        async def on_get(self, req, resp):
            async def emitter():
                yield

            resp.sse = emitter

    resource = SomeResource()

    app = App()
    app.add_route('/', resource)

    client = testing.TestClient(app)

    with pytest.raises(TypeError):
        client.simulate_get()



# Copyright 2013 by Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""HTTP status line constants."""

from typing import Final

# 1xx - Informational
HTTP_100: Final[str] = '100 Continue'
HTTP_CONTINUE: Final[str] = HTTP_100
HTTP_101: Final[str] = '101 Switching Protocols'
HTTP_SWITCHING_PROTOCOLS: Final[str] = HTTP_101
HTTP_102: Final[str] = '102 Processing'
HTTP_PROCESSING: Final[str] = HTTP_102
HTTP_103: Final[str] = '103 Early Hints'
HTTP_EARLY_HINTS: Final[str] = HTTP_103

# 2xx - Success
HTTP_200: Final[str] = '200 OK'
HTTP_OK: Final[str] = HTTP_200
HTTP_201: Final[str] = '201 Created'
HTTP_CREATED: Final[str] = HTTP_201
HTTP_202: Final[str] = '202 Accepted'
HTTP_ACCEPTED: Final[str] = HTTP_202
HTTP_203: Final[str] = '203 Non-Authoritative Information'
HTTP_NON_AUTHORITATIVE_INFORMATION: Final[str] = HTTP_203
HTTP_204: Final[str] = '204 No Content'
HTTP_NO_CONTENT: Final[str] = HTTP_204
HTTP_205: Final[str] = '205 Reset Content'
HTTP_RESET_CONTENT: Final[str] = HTTP_205
HTTP_206: Final[str] = '206 Partial Content'
HTTP_PARTIAL_CONTENT: Final[str] = HTTP_206
HTTP_207: Final[str] = '207 Multi-Status'
HTTP_MULTI_STATUS: Final[str] = HTTP_207
HTTP_208: Final[str] = '208 Already Reported'
HTTP_ALREADY_REPORTED: Final[str] = HTTP_208
HTTP_226: Final[str] = '226 IM Used'
HTTP_IM_USED: Final[str] = HTTP_226

# 3xx - Redirection
HTTP_300: Final[str] = '300 Multiple Choices'
HTTP_MULTIPLE_CHOICES: Final[str] = HTTP_300
HTTP_301: Final[str] = '301 Moved Permanently'
HTTP_MOVED_PERMANENTLY: Final[str] = HTTP_301
HTTP_302: Final[str] = '302 Found'
HTTP_FOUND: Final[str] = HTTP_302
HTTP_303: Final[str] = '303 See Other'
HTTP_SEE_OTHER: Final[str] = HTTP_303
HTTP_304: Final[str] = '304 Not Modified'
HTTP_NOT_MODIFIED: Final[str] = HTTP_304
HTTP_305: Final[str] = '305 Use Proxy'
HTTP_USE_PROXY: Final[str] = HTTP_305
HTTP_307: Final[str] = '307 Temporary Redirect'
HTTP_TEMPORARY_REDIRECT: Final[str] = HTTP_307
HTTP_308: Final[str] = '308 Permanent Redirect'
HTTP_PERMANENT_REDIRECT: Final[str] = HTTP_308

# 4xx - Client Error
HTTP_400: Final[str] = '400 Bad Request'
HTTP_BAD_REQUEST: Final[str] = HTTP_400
HTTP_401: Final[str] = '401 Unauthorized'  # <-- Really means "unauthenticated"
HTTP_UNAUTHORIZED: Final[str] = HTTP_401
HTTP_402: Final[str] = '402 Payment Required'
HTTP_PAYMENT_REQUIRED: Final[str] = HTTP_402
HTTP_403: Final[str] = '403 Forbidden'  # <-- Really means "unauthorized"
HTTP_FORBIDDEN: Final[str] = HTTP_403
HTTP_404: Final[str] = '404 Not Found'
HTTP_NOT_FOUND: Final[str] = HTTP_404
HTTP_405: Final[str] = '405 Method Not Allowed'
HTTP_METHOD_NOT_ALLOWED: Final[str] = HTTP_405
HTTP_406: Final[str] = '406 Not Acceptable'
HTTP_NOT_ACCEPTABLE: Final[str] = HTTP_406
HTTP_407: Final[str] = '407 Proxy Authentication Required'
HTTP_PROXY_AUTHENTICATION_REQUIRED: Final[str] = HTTP_407
HTTP_408: Final[str] = '408 Request Timeout'
HTTP_REQUEST_TIMEOUT: Final[str] = HTTP_408
HTTP_409: Final[str] = '409 Conflict'
HTTP_CONFLICT: Final[str] = HTTP_409
HTTP_410: Final[str] = '410 Gone'
HTTP_GONE: Final[str] = HTTP_410
HTTP_411: Final[str] = '411 Length Required'
HTTP_LENGTH_REQUIRED: Final[str] = HTTP_411
HTTP_412: Final[str] = '412 Precondition Failed'
HTTP_PRECONDITION_FAILED: Final[str] = HTTP_412
HTTP_413: Final[str] = '413 Content Too Large'
HTTP_CONTENT_TOO_LARGE: Final[str] = HTTP_413
HTTP_PAYLOAD_TOO_LARGE: Final[str] = HTTP_413
HTTP_REQUEST_ENTITY_TOO_LARGE: Final[str] = HTTP_413
HTTP_414: Final[str] = '414 URI Too Long'
HTTP_REQUEST_URI_TOO_LONG: Final[str] = HTTP_414
HTTP_415: Final[str] = '415 Unsupported Media Type'
HTTP_UNSUPPORTED_MEDIA_TYPE: Final[str] = HTTP_415
HTTP_416: Final[str] = '416 Range Not Satisfiable'
HTTP_REQUESTED_RANGE_NOT_SATISFIABLE: Final[str] = HTTP_416
HTTP_417: Final[str] = '417 Expectation Failed'
HTTP_EXPECTATION_FAILED: Final[str] = HTTP_417
HTTP_418: Final[str] = "418 I'm a teapot"
HTTP_IM_A_TEAPOT: Final[str] = HTTP_418
HTTP_421: Final[str] = '421 Misdirected Request'
HTTP_MISDIRECTED_REQUEST: Final[str] = HTTP_421
HTTP_422: Final[str] = '422 Unprocessable Entity'
HTTP_UNPROCESSABLE_ENTITY: Final[str] = HTTP_422
HTTP_423: Final[str] = '423 Locked'
HTTP_LOCKED: Final[str] = HTTP_423
HTTP_424: Final[str] = '424 Failed Dependency'
HTTP_FAILED_DEPENDENCY: Final[str] = HTTP_424
HTTP_425: Final[str] = '425 Too Early'
HTTP_TOO_EARLY: Final[str] = HTTP_425
HTTP_426: Final[str] = '426 Upgrade Required'
HTTP_UPGRADE_REQUIRED: Final[str] = HTTP_426
HTTP_428: Final[str] = '428 Precondition Required'
HTTP_PRECONDITION_REQUIRED: Final[str] = HTTP_428
HTTP_429: Final[str] = '429 Too Many Requests'
HTTP_TOO_MANY_REQUESTS: Final[str] = HTTP_429
HTTP_431: Final[str] = '431 Request Header Fields Too Large'
HTTP_REQUEST_HEADER_FIELDS_TOO_LARGE: Final[str] = HTTP_431
HTTP_451: Final[str] = '451 Unavailable For Legal Reasons'
HTTP_UNAVAILABLE_FOR_LEGAL_REASONS: Final[str] = HTTP_451

# 5xx - Server Error
HTTP_500: Final[str] = '500 Internal Server Error'
HTTP_INTERNAL_SERVER_ERROR: Final[str] = HTTP_500
HTTP_501: Final[str] = '501 Not Implemented'
HTTP_NOT_IMPLEMENTED: Final[str] = HTTP_501
HTTP_502: Final[str] = '502 Bad Gateway'
HTTP_BAD_GATEWAY: Final[str] = HTTP_502
HTTP_503: Final[str] = '503 Service Unavailable'
HTTP_SERVICE_UNAVAILABLE: Final[str] = HTTP_503
HTTP_504: Final[str] = '504 Gateway Timeout'
HTTP_GATEWAY_TIMEOUT: Final[str] = HTTP_504
HTTP_505: Final[str] = '505 HTTP Version Not Supported'
HTTP_HTTP_VERSION_NOT_SUPPORTED: Final[str] = HTTP_505
HTTP_506: Final[str] = '506 Variant Also Negotiates'
HTTP_VARIANT_ALSO_NEGOTIATES: Final[str] = HTTP_506
HTTP_507: Final[str] = '507 Insufficient Storage'
HTTP_INSUFFICIENT_STORAGE: Final[str] = HTTP_507
HTTP_508: Final[str] = '508 Loop Detected'
HTTP_LOOP_DETECTED: Final[str] = HTTP_508
HTTP_510: Final[str] = '510 Not Extended'
HTTP_NOT_EXTENDED: Final[str] = HTTP_510
HTTP_511: Final[str] = '511 Network Authentication Required'
HTTP_NETWORK_AUTHENTICATION_REQUIRED: Final[str] = HTTP_511

# 70X - Inexcusable
HTTP_701: Final[str] = '701 Meh'
HTTP_702: Final[str] = '702 Emacs'
HTTP_703: Final[str] = '703 Explosion'

# 71X - Novelty Implementations
HTTP_710: Final[str] = '710 PHP'
HTTP_711: Final[str] = '711 Convenience Store'
HTTP_712: Final[str] = '712 NoSQL'
HTTP_719: Final[str] = '719 I am not a teapot'

# 72X - Edge Cases
HTTP_720: Final[str] = '720 Unpossible'
HTTP_721: Final[str] = '721 Known Unknowns'
HTTP_722: Final[str] = '722 Unknown Unknowns'
HTTP_723: Final[str] = '723 Tricky'
HTTP_724: Final[str] = '724 This line should be unreachable'
HTTP_725: Final[str] = '725 It works on my machine'
HTTP_726: Final[str] = "726 It's a feature, not a bug"
HTTP_727: Final[str] = '727 32 bits is plenty'

# 74X - Meme Driven
HTTP_740: Final[str] = '740 Computer says no'
HTTP_741: Final[str] = '741 Compiling'
HTTP_742: Final[str] = '742 A kitten dies'
HTTP_743: Final[str] = '743 I thought I knew regular expressions'
HTTP_744: Final[str] = '744 Y U NO write integration tests?'
HTTP_745: Final[str] = (
    "745 I don't always test my code, but when I do I do it in production"
)
HTTP_748: Final[str] = '748 Confounded by Ponies'
HTTP_749: Final[str] = '749 Reserved for Chuck Norris'

# 75X - Syntax Errors
HTTP_750: Final[str] = "750 Didn't bother to compile it"
HTTP_753: Final[str] = '753 Syntax Error'
HTTP_754: Final[str] = '754 Too many semi-colons'
HTTP_755: Final[str] = '755 Not enough semi-colons'
HTTP_759: Final[str] = '759 Unexpected T_PAAMAYIM_NEKUDOTAYIM'

# 77X - Predictable Problems
HTTP_771: Final[str] = '771 Cached for too long'
HTTP_772: Final[str] = '772 Not cached long enough'
HTTP_773: Final[str] = '773 Not cached at all'
HTTP_774: Final[str] = '774 Why was this cached?'
HTTP_776: Final[str] = '776 Error on the Exception'
HTTP_777: Final[str] = '777 Coincidence'
HTTP_778: Final[str] = '778 Off By One Error'
HTTP_779: Final[str] = '779 Off By Too Many To Count Error'

# 78X - Somebody Else's Problem
HTTP_780: Final[str] = '780 Project owner not responding'
HTTP_781: Final[str] = '781 Operations'
HTTP_782: Final[str] = '782 QA'
HTTP_783: Final[str] = '783 It was a customer request, honestly'
HTTP_784: Final[str] = '784 Management, obviously'
HTTP_785: Final[str] = '785 TPS Cover Sheet not attached'
HTTP_786: Final[str] = '786 Try it now'

# 79X - Internet crashed
HTTP_791: Final[str] = '791 The Internet shut down due to copyright restrictions'
HTTP_792: Final[str] = '792 Climate change driven catastrophic weather event'
HTTP_797: Final[str] = '797 This is the last page of the Internet. Go back'
HTTP_799: Final[str] = '799 End of the world'

__all__ = (
    'HTTP_100',
    'HTTP_101',
    'HTTP_102',
    'HTTP_103',
    'HTTP_200',
    'HTTP_201',
    'HTTP_202',
    'HTTP_203',
    'HTTP_204',
    'HTTP_205',
    'HTTP_206',
    'HTTP_207',
    'HTTP_208',
    'HTTP_226',
    'HTTP_300',
    'HTTP_301',
    'HTTP_302',
    'HTTP_303',
    'HTTP_304',
    'HTTP_305',
    'HTTP_307',
    'HTTP_308',
    'HTTP_400',
    'HTTP_401',
    'HTTP_402',
    'HTTP_403',
    'HTTP_404',
    'HTTP_405',
    'HTTP_406',
    'HTTP_407',
    'HTTP_408',
    'HTTP_409',
    'HTTP_410',
    'HTTP_411',
    'HTTP_412',
    'HTTP_413',
    'HTTP_414',
    'HTTP_415',
    'HTTP_416',
    'HTTP_417',
    'HTTP_418',
    'HTTP_421',
    'HTTP_422',
    'HTTP_423',
    'HTTP_424',
    'HTTP_425',
    'HTTP_426',
    'HTTP_428',
    'HTTP_429',
    'HTTP_431',
    'HTTP_451',
    'HTTP_500',
    'HTTP_501',
    'HTTP_502',
    'HTTP_503',
    'HTTP_504',
    'HTTP_505',
    'HTTP_506',
    'HTTP_507',
    'HTTP_508',
    'HTTP_510',
    'HTTP_511',
    'HTTP_701',
    'HTTP_702',
    'HTTP_703',
    'HTTP_710',
    'HTTP_711',
    'HTTP_712',
    'HTTP_719',
    'HTTP_720',
    'HTTP_721',
    'HTTP_722',
    'HTTP_723',
    'HTTP_724',
    'HTTP_725',
    'HTTP_726',
    'HTTP_727',
    'HTTP_740',
    'HTTP_741',
    'HTTP_742',
    'HTTP_743',
    'HTTP_744',
    'HTTP_745',
    'HTTP_748',
    'HTTP_749',
    'HTTP_750',
    'HTTP_753',
    'HTTP_754',
    'HTTP_755',
    'HTTP_759',
    'HTTP_771',
    'HTTP_772',
    'HTTP_773',
    'HTTP_774',
    'HTTP_776',
    'HTTP_777',
    'HTTP_778',
    'HTTP_779',
    'HTTP_780',
    'HTTP_781',
    'HTTP_782',
    'HTTP_783',
    'HTTP_784',
    'HTTP_785',
    'HTTP_786',
    'HTTP_791',
    'HTTP_792',
    'HTTP_797',
    'HTTP_799',
    'HTTP_ACCEPTED',
    'HTTP_ALREADY_REPORTED',
    'HTTP_BAD_GATEWAY',
    'HTTP_BAD_REQUEST',
    'HTTP_CONFLICT',
    'HTTP_CONTENT_TOO_LARGE',
    'HTTP_CONTINUE',
    'HTTP_CREATED',
    'HTTP_EARLY_HINTS',
    'HTTP_EXPECTATION_FAILED',
    'HTTP_FAILED_DEPENDENCY',
    'HTTP_FORBIDDEN',
    'HTTP_FOUND',
    'HTTP_GATEWAY_TIMEOUT',
    'HTTP_GONE',
    'HTTP_HTTP_VERSION_NOT_SUPPORTED',
    'HTTP_IM_A_TEAPOT',
    'HTTP_IM_USED',
    'HTTP_INSUFFICIENT_STORAGE',
    'HTTP_INTERNAL_SERVER_ERROR',
    'HTTP_LENGTH_REQUIRED',
    'HTTP_LOCKED',
    'HTTP_LOOP_DETECTED',
    'HTTP_METHOD_NOT_ALLOWED',
    'HTTP_MISDIRECTED_REQUEST',
    'HTTP_MOVED_PERMANENTLY',
    'HTTP_MULTIPLE_CHOICES',
    'HTTP_MULTI_STATUS',
    'HTTP_NETWORK_AUTHENTICATION_REQUIRED',
    'HTTP_NON_AUTHORITATIVE_INFORMATION',
    'HTTP_NOT_ACCEPTABLE',
    'HTTP_NOT_EXTENDED',
    'HTTP_NOT_FOUND',
    'HTTP_NOT_IMPLEMENTED',
    'HTTP_NOT_MODIFIED',
    'HTTP_NO_CONTENT',
    'HTTP_OK',
    'HTTP_PARTIAL_CONTENT',
    'HTTP_PAYMENT_REQUIRED',
    'HTTP_PERMANENT_REDIRECT',
    'HTTP_PRECONDITION_FAILED',
    'HTTP_PRECONDITION_REQUIRED',
    'HTTP_PROCESSING',
    'HTTP_PROXY_AUTHENTICATION_REQUIRED',
    'HTTP_REQUESTED_RANGE_NOT_SATISFIABLE',
    'HTTP_REQUEST_HEADER_FIELDS_TOO_LARGE',
    'HTTP_REQUEST_TIMEOUT',
    'HTTP_REQUEST_URI_TOO_LONG',
    'HTTP_RESET_CONTENT',
    'HTTP_SEE_OTHER',
    'HTTP_SERVICE_UNAVAILABLE',
    'HTTP_SWITCHING_PROTOCOLS',
    'HTTP_TEMPORARY_REDIRECT',
    'HTTP_TOO_EARLY',
    'HTTP_TOO_MANY_REQUESTS',
    'HTTP_UNAUTHORIZED',
    'HTTP_UNAVAILABLE_FOR_LEGAL_REASONS',
    'HTTP_UNPROCESSABLE_ENTITY',
    'HTTP_UNSUPPORTED_MEDIA_TYPE',
    'HTTP_UPGRADE_REQUIRED',
    'HTTP_USE_PROXY',
    'HTTP_VARIANT_ALSO_NEGOTIATES',
)



from __future__ import annotations

from functools import partial
import json
from typing import Any, Callable, Optional, Tuple, Type, Union

from falcon import errors
from falcon import http_error
from falcon.media.base import BaseHandler
from falcon.media.base import TextBaseHandlerWS
from falcon.typing import AsyncReadableIO
from falcon.typing import ReadableIO


class JSONHandler(BaseHandler):
    """JSON media handler.

    This handler uses Python's standard :mod:`json` library by default, but
    can be easily configured to use any of a number of third-party JSON
    libraries, depending on your needs. For example, you can often
    realize a significant performance boost under CPython by using an
    alternative library. Good options in this respect include
    :ref:`msgspec <msgspec_recipe>`, `orjson`, `python-rapidjson`, and `mujson`.

    This handler will raise a :class:`falcon.MediaNotFoundError` when attempting
    to parse an empty body, or a :class:`falcon.MediaMalformedError`
    if an error happens while parsing the body.

    Note:
        If you are deploying to PyPy, we recommend sticking with the standard
        library's JSON implementation, since it will be faster in most cases
        as compared to a third-party library.

    .. _custom-media-json-library:

    .. rubric:: Custom JSON library

    You can replace the default JSON handler by using a custom JSON library
    (see also: :ref:`custom_media_handlers`). Overriding the default JSON
    implementation is simply a matter of specifying the desired ``dumps`` and
    ``loads`` functions::

        import falcon
        from falcon import media

        import rapidjson

        json_handler = media.JSONHandler(
            dumps=rapidjson.dumps,
            loads=rapidjson.loads,
        )
        extra_handlers = {
            'application/json': json_handler,
        }

        app = falcon.App()
        app.req_options.media_handlers.update(extra_handlers)
        app.resp_options.media_handlers.update(extra_handlers)

    .. rubric:: Custom serialization parameters

    Even if you decide to stick with the stdlib's :any:`json.dumps` and
    :any:`json.loads`, you can wrap them using :any:`functools.partial` to
    provide custom serialization or deserialization parameters supported by the
    ``dumps`` and ``loads`` functions, respectively
    (see also: :ref:`prettifying-json-responses`)::

        import falcon
        from falcon import media

        from functools import partial

        json_handler = media.JSONHandler(
            dumps=partial(
                json.dumps,
                default=str,
                sort_keys=True,
            ),
        )
        extra_handlers = {
            'application/json': json_handler,
        }

        app = falcon.App()
        app.req_options.media_handlers.update(extra_handlers)
        app.resp_options.media_handlers.update(extra_handlers)

    By default, ``ensure_ascii`` is passed to the ``json.dumps`` function.
    If you override the ``dumps`` function, you will need to explicitly set
    ``ensure_ascii`` to ``False`` in order to enable the serialization of
    Unicode characters to UTF-8. This is easily done by using
    :any:`functools.partial` to apply the desired keyword argument. As also
    demonstrated in the previous paragraph, you can use this same technique to
    customize any option supported by the ``dumps`` and ``loads`` functions::

        from functools import partial

        from falcon import media
        import rapidjson

        json_handler = media.JSONHandler(
            dumps=partial(
                rapidjson.dumps,
                ensure_ascii=False, sort_keys=True
            ),
        )

    .. _custom-media-json-encoder:

    .. rubric:: Custom JSON encoder

    You can also override the default :class:`~json.JSONEncoder` by using a
    custom Encoder and updating the media handlers for ``application/json``
    type to use that::

        import json
        from datetime import datetime
        from functools import partial

        import falcon
        from falcon import media

        class DatetimeEncoder(json.JSONEncoder):
            \"\"\"Json Encoder that supports datetime objects.\"\"\"

            def default(self, obj):
                if isinstance(obj, datetime):
                    return obj.isoformat()
                return super().default(obj)

        app = falcon.App()

        json_handler = media.JSONHandler(
            dumps=partial(json.dumps, cls=DatetimeEncoder),
        )
        extra_handlers = {
            'application/json': json_handler,
        }

        app.req_options.media_handlers.update(extra_handlers)
        app.resp_options.media_handlers.update(extra_handlers)

    .. note:: When testing an application employing a custom JSON encoder, bear
        in mind that :class:`~.testing.TestClient` is decoupled from the app,
        and it simulates requests as if they were performed by a third-party
        client (just sans network). Therefore, passing the **json** parameter
        to :ref:`simulate_* <testing_standalone_methods>` methods will
        effectively use the stdlib's :func:`json.dumps`. If you want to
        serialize custom objects for testing, you will need to dump them into a
        string yourself, and pass it using the **body** parameter instead
        (accompanied by the ``application/json`` content type header).

    Keyword Arguments:
        dumps (func): Function to use when serializing JSON responses.
        loads (func): Function to use when deserializing JSON requests.
    """

    def __init__(
        self,
        dumps: Optional[Callable[[Any], Union[str, bytes]]] = None,
        loads: Optional[Callable[[str], Any]] = None,
    ) -> None:
        self._dumps = dumps or partial(json.dumps, ensure_ascii=False)
        self._loads = loads or json.loads
        self._deserialization_errors: Tuple[Type[Exception], ...] = (ValueError,)

        # PERF(kgriffs): Test dumps once up front so we can set the
        #     proper serialize implementation.
        result = self._dumps({'message': 'Hello World'})
        if isinstance(result, str):
            self.serialize = self._serialize_s  # type: ignore[method-assign]
            self.serialize_async = self._serialize_async_s  # type: ignore[method-assign]
        else:
            self.serialize = self._serialize_b  # type: ignore[method-assign]
            self.serialize_async = self._serialize_async_b  # type: ignore[method-assign]

        # PERF(vytas): Try to detect nonstandard deserialization exception
        #   classes (that do not subclass ValueError) up front, and add them to
        #   the tuple of known deserialization errors.
        try:
            self._loads('frozenset({1})')
        except Exception as ex:
            error_cls = type(ex)
            if not issubclass(error_cls, ValueError):
                self._deserialization_errors += (error_cls,)

        # NOTE(kgriffs): To be safe, only enable the optimized protocol when
        #   not subclassed.
        if type(self) is JSONHandler:
            self._serialize_sync = self.serialize
            self._deserialize_sync = self._deserialize

    def _deserialize(self, data: bytes) -> Any:
        if not data:
            raise errors.MediaNotFoundError('JSON')
        try:
            return self._loads(data.decode())
        except self._deserialization_errors as ex:
            raise errors.MediaMalformedError('JSON') from ex

    def deserialize(
        self,
        stream: ReadableIO,
        content_type: Optional[str],
        content_length: Optional[int],
    ) -> Any:
        return self._deserialize(stream.read())

    async def deserialize_async(
        self,
        stream: AsyncReadableIO,
        content_type: Optional[str],
        content_length: Optional[int],
    ) -> Any:
        return self._deserialize(await stream.read())

    # NOTE(kgriffs): Make content_type a kwarg to support the
    #   Request.render_body() shortcut optimization.
    def _serialize_s(self, media: Any, content_type: Optional[str] = None) -> bytes:
        return self._dumps(media).encode()  # type: ignore[union-attr]

    async def _serialize_async_s(
        self, media: Any, content_type: Optional[str]
    ) -> bytes:
        return self._dumps(media).encode()  # type: ignore[union-attr]

    # NOTE(kgriffs): Make content_type a kwarg to support the
    #   Request.render_body() shortcut optimization.
    def _serialize_b(self, media: Any, content_type: Optional[str] = None) -> bytes:
        return self._dumps(media)  # type: ignore[return-value]

    async def _serialize_async_b(
        self, media: Any, content_type: Optional[str]
    ) -> bytes:
        return self._dumps(media)  # type: ignore[return-value]


class JSONHandlerWS(TextBaseHandlerWS):
    """WebSocket media handler for de(serializing) JSON to/from TEXT payloads.

    This handler uses Python's standard :mod:`json` library by default, but
    can be easily configured to use any of a number of third-party JSON
    libraries, depending on your needs. For example, you can often
    realize a significant performance boost under CPython by using an
    alternative library. Good options in this respect include `orjson`,
    `python-rapidjson`, and `mujson`.

    Note:
        If you are deploying to PyPy, we recommend sticking with the standard
        library's JSON implementation, since it will be faster in most cases
        as compared to a third-party library.

    Overriding the default JSON implementation is simply a matter of specifying
    the desired ``dumps`` and ``loads`` functions::

        import falcon
        from falcon import media

        import rapidjson

        json_handler = media.JSONHandlerWS(
            dumps=rapidjson.dumps,
            loads=rapidjson.loads,
        )

        app = falcon.asgi.App()
        app.ws_options.media_handlers[falcon.WebSocketPayloadType.TEXT] = json_handler

    By default, ``ensure_ascii`` is passed to the ``json.dumps`` function.
    If you override the ``dumps`` function, you will need to explicitly set
    ``ensure_ascii`` to ``False`` in order to enable the serialization of
    Unicode characters to UTF-8. This is easily done by using
    :any:`functools.partial` to apply the desired keyword argument. In fact, you
    can use this same technique to customize any option supported by the
    ``dumps`` and ``loads`` functions::

        from functools import partial

        from falcon import media
        import rapidjson

        json_handler = media.JSONHandlerWS(
            dumps=partial(
                rapidjson.dumps,
                ensure_ascii=False, sort_keys=True
            ),
        )

    Keyword Arguments:
        dumps (func): Function to use when serializing JSON.
        loads (func): Function to use when deserializing JSON.
    """

    __slots__ = ['dumps', 'loads']

    def __init__(
        self,
        dumps: Optional[Callable[[Any], str]] = None,
        loads: Optional[Callable[[str], Any]] = None,
    ) -> None:
        self._dumps = dumps or partial(json.dumps, ensure_ascii=False)
        self._loads = loads or json.loads

    def serialize(self, media: object) -> str:
        return self._dumps(media)

    def deserialize(self, payload: str) -> object:
        return self._loads(payload)


http_error._DEFAULT_JSON_HANDLER = _DEFAULT_JSON_HANDLER = JSONHandler()



import io

import pytest

from falcon.errors import DelimiterError
from falcon.util import BufferedReader


class WouldHang(RuntimeError):
    pass


class GlitchyStream(io.BytesIO):
    def read(self, size=None):
        if size is None or size == -1:
            raise WouldHang('unbounded read()')

        result = super().read(size)
        if not result:
            raise WouldHang('EOF')
        return result


class FragmentedStream(GlitchyStream):
    def read(self, size=None):
        if size is None or size <= 1:
            return super().read(size)

        if size < 8:
            size = 1
        else:
            size = size // 2

        return super().read(size)


TEST_DATA = (
    b'123456789ABCDEF\n' * 64 * 1024 * 64
    + b'--boundary1234567890--'
    + b'123456789ABCDEF\n' * 64 * 1024 * 63
    + b'--boundary1234567890--'
    + b'123456789ABCDEF\n' * 64 * 1024 * 62
    + b'--boundary1234567890--'
)

TEST_BYTES_IO = GlitchyStream(TEST_DATA)


@pytest.fixture()
def buffered_reader():
    def _reader_fixture(chunk_size=None):
        TEST_BYTES_IO.seek(0)
        return BufferedReader(TEST_BYTES_IO.read, len(TEST_DATA), chunk_size)

    yield _reader_fixture


@pytest.fixture()
def shorter_stream():
    TEST_BYTES_IO.seek(0)
    return BufferedReader(TEST_BYTES_IO.read, 1024, 128)


@pytest.fixture()
def fragmented_stream():
    stream = FragmentedStream(TEST_DATA)
    return BufferedReader(stream.read, len(TEST_DATA))


def test_peek(buffered_reader):
    stream = buffered_reader(16)

    assert stream.peek(0) == b''
    assert stream.peek(1) == b'1'
    assert stream.peek(2) == b'12'
    assert stream.peek(16) == b'123456789ABCDEF\n'
    assert stream.peek(17) == b'123456789ABCDEF\n'
    assert stream.peek(32) == b'123456789ABCDEF\n'

    assert stream.read(15) == b'123456789ABCDEF'

    assert stream.peek(0) == b''
    assert stream.peek(1) == b'\n'
    assert stream.peek(2) == b'\n1'
    assert stream.peek(16) == b'\n123456789ABCDEF'
    assert stream.peek(17) == b'\n123456789ABCDEF'
    assert stream.peek(32) == b'\n123456789ABCDEF'


def test_peek_eof():
    source = b'Hello, world!\n'
    source_stream = io.BytesIO(source)
    stream = BufferedReader(source_stream.read, len(source) - 1)

    assert stream.peek(0) == b''
    assert stream.peek(1) == b'H'
    assert stream.peek(2) == b'He'
    assert stream.peek(16) == b'Hello, world!'
    assert stream.peek(32) == b'Hello, world!'

    assert source_stream.read() == b'\n'


def test_bounded_read():
    stream = io.BytesIO(b'Hello, world!')
    buffered = BufferedReader(stream.read, len('Hello, world'))
    buffered.read()

    assert stream.read() == b'!'


@pytest.mark.parametrize(
    'size',
    [
        0,
        1,
        2,
        7,
        62,
        63,
        64,
        65,
        126,
        127,
        128,
        129,
        1000,
        10000,
    ],
)
def test_read_from_buffer(buffered_reader, size):
    stream = buffered_reader(64)
    stream.peek(64)

    assert stream.read(size) == TEST_DATA[:size]
    assert stream.read(1) == TEST_DATA[size : size + 1]


def test_read_until_delimiter_size_check(buffered_reader):
    stream = buffered_reader(64)

    with pytest.raises(ValueError):
        stream.read_until(b'')
    with pytest.raises(ValueError):
        stream.read_until(b'B' * 65)


@pytest.mark.parametrize(
    'size',
    [
        0,
        1,
        2,
        7,
        62,
        63,
        64,
        65,
        126,
        127,
        128,
        129,
        1000,
        10000,
    ],
)
def test_read_until_with_size(buffered_reader, size):
    stream = buffered_reader(64)
    assert stream.read_until(b'--boundary1234567890--', size) == (TEST_DATA[:size])


@pytest.mark.slow
def test_read_until(buffered_reader):
    stream = buffered_reader()

    assert len(stream.read_until(b'--boundary1234567890--')) == 64 * 1024**2
    stream.read_until(b'123456789ABCDEF\n')
    assert len(stream.read_until(b'--boundary1234567890--')) == 63 * 1024**2
    stream.read_until(b'123456789ABCDEF\n')
    assert len(stream.read_until(b'--boundary1234567890--')) == 62 * 1024**2


@pytest.mark.parametrize(
    'size1,size2',
    [
        (11003077, 22000721),
        (13372477, 51637898),
    ],
)
@pytest.mark.slow
def test_irregular_large_read_until(buffered_reader, size1, size2):
    stream = buffered_reader()
    delimiter = b'--boundary1234567890--'

    stream.pipe_until(delimiter, consume_delimiter=True)
    stream.pipe_until(delimiter, consume_delimiter=True)

    expected = b'123456789ABCDEF\n' * 64 * 1024 * 62

    assert stream.read_until(delimiter, 1337) == expected[:1337]

    chunk1 = stream.read_until(delimiter, size1)
    assert len(chunk1) == size1

    chunk2 = stream.read_until(delimiter, size2)
    assert len(chunk2) == size2

    remainder = stream.read_until(delimiter, 62 * 1024 * 1024)
    assert chunk1 + chunk2 + remainder == expected[1337:]


@pytest.mark.parametrize(
    'size',
    [
        0,
        1,
        2,
        7,
        62,
        63,
        64,
        65,
        126,
        127,
        128,
        129,
        1000,
    ],
)
def test_read_until_from_buffer(shorter_stream, size):
    shorter_stream.peek(128)

    assert shorter_stream.read_until(b'\n1', size) == (b'123456789ABCDEF'[:size])


def test_read_until_missing_delimiter(shorter_stream):
    class BoundaryMissing(RuntimeError):
        pass

    with pytest.raises(DelimiterError):
        shorter_stream.read_until(b'--boundary1234567890--', consume_delimiter=True)


def test_consume_delimiter(shorter_stream):
    shorter_stream.peek()
    shorter_stream.read(113)

    assert shorter_stream.peek(1) == b'2'

    chunk = shorter_stream.read_until(b'1', 15, consume_delimiter=True)
    assert chunk == b'23456789ABCDEF\n'

    assert shorter_stream.read_until(b'ABCDE', 4) == b'2345'


@pytest.mark.parametrize('chunk_size', list(range(46, 63)))
def test_read_until_shared_boundary(chunk_size):
    source = b'-boundary-like-' * 4 + b'--some junk--\n' + b'\n' * 1024
    source_stream = io.BytesIO(source)
    stream = BufferedReader(source_stream.read, len(source), chunk_size)
    assert stream.read_until(b'-boundary-like---') == b'-boundary-like-' * 3
    assert stream.peek(17) == b'-boundary-like---'


def test_pipe(shorter_stream):
    output = io.BytesIO()
    shorter_stream.pipe(output)
    assert len(output.getvalue()) == 1024


def test_pipe_until(buffered_reader):
    stream = buffered_reader(2**16)

    output = io.BytesIO()
    stream.pipe_until(b'--boundary1234567890--', output)
    assert len(output.getvalue()) == 64 * 1024**2


def test_pipe_until_without_destination(buffered_reader):
    stream = buffered_reader(2**16)
    stream.pipe_until(b'--boundary1234567890--')
    assert stream.peek(22) == b'--boundary1234567890--'


def test_exhaust(shorter_stream):
    shorter_stream.exhaust()
    assert not shorter_stream.read()


def test_readline():
    source = (
        b'Hello, world!\nA line.\n\nA longer line... \n' + b'SPAM ' * 7 + b'\n' + b'\n'
    )

    stream = BufferedReader(io.BytesIO(source).read, len(source))
    assert stream.readline() == b'Hello, world!\n'
    assert stream.readline() == b'A line.\n'
    assert stream.readline() == b'\n'
    assert stream.readline() == b'A longer line... \n'
    assert stream.readline() == b'SPAM SPAM SPAM SPAM SPAM SPAM SPAM \n'
    assert stream.readline() == b'\n'
    assert stream.readline() == b''
    assert stream.readline() == b''


def test_readline_with_size():
    source = (
        b'Hello, world! This is a short village name in Wales.\n'
        b'Llanfairpwllgwyngyllgogerychwyrndrobwllllantysiliogogogoch'
    )

    stream = BufferedReader(io.BytesIO(source).read, len(source))
    assert stream.readline(37) == b'Hello, world! This is a short village'
    assert stream.readline(37) == b' name in Wales.\n'
    assert stream.readline(8) == b'Llanfair'
    assert stream.readline(16) == b'pwllgwyngyllgoge'
    assert stream.readline(64) == b'rychwyrndrobwllllantysiliogogogoch'


def test_readlines(shorter_stream):
    assert shorter_stream.readlines() == [b'123456789ABCDEF\n'] * 64


@pytest.mark.parametrize(
    'chunk_size',
    [
        8,
        16,
        256,
        1024,
        2**16,
    ],
)
def test_readlines_hint(buffered_reader, chunk_size):
    stream = buffered_reader(chunk_size)
    assert stream.readlines(100) == [b'123456789ABCDEF\n'] * 7
    assert stream.readlines(64) == [b'123456789ABCDEF\n'] * 4
    assert stream.readlines(16) == [b'123456789ABCDEF\n']


def test_duck_compatibility_with_io_base(shorter_stream):
    assert shorter_stream.readable()
    assert not shorter_stream.seekable()
    assert not shorter_stream.writeable()


@pytest.mark.slow
def test_fragmented_reads(fragmented_stream):
    b = io.BytesIO()
    fragmented_stream.pipe_until(b'--boundary1234567890--', b)
    assert fragmented_stream.read(2) == b'--'

    fragmented_stream.pipe_until(b'--boundary1234567890--')
    assert fragmented_stream.read(3) == b'--b'

    fragmented_stream.pipe_until(b'--boundary1234567890--')
    assert fragmented_stream.read(7) == b'--bound'

    fragmented_stream.exhaust()
    assert fragmented_stream.read(4) == b''
    assert fragmented_stream.read() == b''



import falcon
import falcon.media


class Forms:
    def on_post(self, req, resp):
        example = {}
        for part in req.media:
            if part.content_type.startswith('multipart/mixed'):
                for nested in part.media:
                    example[nested.filename] = nested.text

        resp.media = example


parser = falcon.media.MultipartFormHandler()
parser.parse_options.media_handlers['multipart/mixed'] = (
    falcon.media.MultipartFormHandler()
)

app = falcon.App()
app.req_options.media_handlers[falcon.MEDIA_MULTIPART] = parser
app.add_route('/forms', Forms())



import inspect

from falcon.asgi import _asgi_helpers


class IntricateApp:
    def unorthodox_call(self, scope, receive, send):
        return self._call_factory(scope, receive, send)

    async def _call_factory(self, scope, receive, send):
        await send('Hello!')
        await send('Bye.')

    __call__ = _asgi_helpers._wrap_asgi_coroutine_func(unorthodox_call)


async def test_intricate_app():
    async def receive():
        pass

    async def send(msg):
        messages.append(msg)

    app = IntricateApp()
    messages = []

    assert not inspect.iscoroutinefunction(app.unorthodox_call)
    assert inspect.iscoroutinefunction(app.__call__)

    await app({}, receive, send)

    assert messages == ['Hello!', 'Bye.']



import pytest

import falcon
from falcon import testing
from falcon.asgi import App


async def test_default_headers():
    class Resource:
        async def on_get(self, req, resp):
            resp.set_header('the-answer', req.get_header('the-answer'))

    app = App()
    app.add_route('/', Resource())
    client = testing.TestClient(app, headers={'the-answer': '42'})

    async with client as conductor:
        result = await conductor.simulate_get()
        assert result.headers['the-answer'] == '42'


@pytest.mark.parametrize('simulate_method', ['request', 'simulate_request'])
async def test_generic_request(simulate_method):
    class Resource:
        async def on_lock(self, req, resp):
            raise falcon.HTTPUnprocessableEntity

    app = App()
    app.add_route('/file', Resource())
    client = testing.TestClient(app)

    async with client as conductor:
        simulate_request = getattr(conductor, simulate_method)
        result = await simulate_request('LOCK', '/file')
        assert result.status_code == 422


async def test_wsgi_not_supported():
    with pytest.raises(falcon.CompatibilityError):
        async with testing.TestClient(falcon.App()):
            pass

    with pytest.raises(falcon.CompatibilityError):
        async with testing.ASGIConductor(falcon.App()):
            pass


@pytest.mark.parametrize(
    'method', ['get', 'head', 'post', 'put', 'options', 'patch', 'delete']
)
@pytest.mark.parametrize('use_alias', ['alias', 'simulate'])
async def test_responders(method, use_alias):
    class Resource:
        async def on_get(self, req, resp):
            resp.set_header('method', 'get')

        async def on_head(self, req, resp):
            resp.set_header('method', 'head')

        async def on_post(self, req, resp):
            resp.set_header('method', 'post')

        async def on_put(self, req, resp):
            resp.set_header('method', 'put')

        async def on_options(self, req, resp):
            resp.set_header('method', 'options')

        async def on_patch(self, req, resp):
            resp.set_header('method', 'patch')

        async def on_delete(self, req, resp):
            resp.set_header('method', 'delete')

    resource = Resource()

    app = App()
    app.add_route('/', resource)

    async with testing.ASGIConductor(app) as conductor:
        name = method if use_alias == 'alias' else 'simulate_' + method
        simulate = getattr(conductor, name)
        result = await simulate('/')
        assert result.headers['method'] == method
